from unittest import mock

import pytest

from mirrors_countme.scripts.countme_totals import cli, parse_args


# Test parse_args
def test_parse_args_with_required_arguments():
    argv = ["countme.db"]
    args = parse_args(argv)
    assert args.countme_totals == "countme.db"


def test_parse_args_with_optional_arguments():
    argv = ["countme.db", "--update-from", "raw.db", "--csv-dump", "output.csv", "--progress"]
    args = parse_args(argv)
    assert args.countme_totals == "countme.db"
    assert args.countme_raw == "raw.db"
    assert args.csv_dump.name == "output.csv"
    assert args.progress


def test_parse_args_with_version_argument():
    argv = ["-V"]
    with pytest.raises(SystemExit) as exc_info:
        parse_args(argv)

    assert exc_info.value.code == 0


def test_parse_args_with_missing_required_argument():
    argv = ["--update-from", "raw.db", "--csv-dump", "output.csv", "--progress"]
    with pytest.raises(SystemExit) as exc_info:
        parse_args(argv)

    assert exc_info.value.code == 2


# Test CLI
@mock.patch("mirrors_countme.scripts.countme_totals.parse_args")
@mock.patch("mirrors_countme.scripts.countme_totals.totals")
def test_cli(mock_totals, mock_parse_args):
    mock_args = mock.Mock(
        countme_totals="countme.db",
        countme_raw="",
        progress="--progress",
        csv_dump="",
    )
    mock_parse_args.return_value = mock_args

    cli()

    mock_parse_args.assert_called_once()

    mock_totals.assert_called_once_with(
        countme_totals="countme.db", countme_raw="", progress="--progress", csv_dump=""
    )


@mock.patch("mirrors_countme.scripts.countme_totals.parse_args")
def test_cli_keyboard_interrupt(mock_parse_args):
    mock_parse_args.side_effect = KeyboardInterrupt

    with pytest.raises(SystemExit) as exc_info:
        cli()
    assert exc_info.value.code == 3



import locale
import os
from pathlib import Path

import pytest

if [int(x) for x in pytest.__version__.split(".")] < [3, 9, 0]:
    # the tmp_path fixture is only available in pytest >= 3.9.0
    @pytest.fixture
    def tmp_path(tmpdir):
        return Path(tmpdir)


@pytest.fixture
def tmp_path_cwd(tmp_path):
    """Return a temporary path and change into it"""
    old_wd = os.getcwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(old_wd)


@pytest.fixture(autouse=True)
def use_english_locale():
    locale.setlocale(locale.LC_ALL, "en_US.UTF-8")



#!/usr/bin/python3

import argparse

from ..totals import totals
from ..version import __version__

# ===========================================================================
# ====== CLI parser & __main__ ==============================================
# ===========================================================================


def parse_args(argv=None):
    p = argparse.ArgumentParser(
        description="Aggregate 'countme' log records to weekly totals.",
    )
    p.add_argument("-V", "--version", action="version", version=f"%(prog)s {__version__}")

    p.add_argument("countme_totals", help="Database containing countme_totals")

    p.add_argument(
        "--update-from",
        metavar="COUNTME_RAW_DB",
        dest="countme_raw",
        help="Update totals from raw data (from countme_parse_access_log)",
    )

    p.add_argument(
        "--csv-dump",
        type=argparse.FileType("wt", encoding="utf-8"),
        help="File to dump CSV-formatted totals data",
    )

    p.add_argument(
        "--progress",
        action="store_true",
        help="Show progress while reading and counting data.",
    )

    args = p.parse_args(argv)

    return args


def cli():
    try:
        args = parse_args()
        totals(
            countme_totals=args.countme_totals,
            countme_raw=args.countme_raw,
            progress=args.progress,
            csv_dump=args.csv_dump,
        )
    except KeyboardInterrupt:
        raise SystemExit(3)  # You know, 3, like 'C', like Ctrl-C!



import datetime as dt
from contextlib import nullcontext
from unittest import mock

import pytest
from hypothesis import given
from hypothesis.strategies import integers

from mirrors_countme import constants
from mirrors_countme.scripts import countme_trim_raw
from mirrors_countme.version import __version__

from ..common import MAX_TIMESTAMP

NOW_TIMESTAMP = int(dt.datetime.utcnow().timestamp())


@pytest.mark.parametrize(
    "value, expected",
    (
        ("1", 1),
        ("0", ValueError),
        ("-1", ValueError),
        ("boop", ValueError),
    ),
)
def test_positive_int(value, expected):
    if isinstance(expected, type) and issubclass(expected, Exception):
        with pytest.raises(expected):
            countme_trim_raw.positive_int(value)
    else:
        assert countme_trim_raw.positive_int(value) == expected


class TestParseArgs:
    def test_help(self, capsys):
        with pytest.raises(SystemExit) as exc_info:
            countme_trim_raw.parse_args(["--help"])
        assert exc_info.value.code == 0
        stdout, stderr = capsys.readouterr()
        assert stdout.startswith("usage:")
        assert not stderr

    def test_version(self, capsys):
        with pytest.raises(SystemExit) as exc_info:
            countme_trim_raw.parse_args(["--version"])
        assert exc_info.value.code == 0
        stdout, stderr = capsys.readouterr()
        assert __version__ in stdout
        assert not stderr

    def test_dbfile_missing(self, capsys):
        with pytest.raises(SystemExit) as exc_info:
            countme_trim_raw.parse_args([])
        assert exc_info.value.code != 0
        stdout, stderr = capsys.readouterr()
        assert not stdout
        assert "error: the following arguments are required" in stderr

    @pytest.mark.parametrize(
        "value, expected",
        (
            (None, False),
            ("--read-write", True),
            ("--noop", False),
        ),
    )
    def test_rw(self, value, expected):
        argv = ["test.db"]
        if value:
            argv.append(value)
        _, args = countme_trim_raw.parse_args(argv)
        assert args.rw == expected

    @pytest.mark.parametrize(
        "value, expected",
        (
            (None, False),
            ("--oldest-week", True),
        ),
    )
    def test_oldest_week(self, value, expected):
        argv = ["test.db"]
        if value:
            argv.append(value)
        _, args = countme_trim_raw.parse_args(argv)
        assert args.oldest_week == expected

    @pytest.mark.parametrize(
        "value, expected",
        (
            (None, countme_trim_raw.CONF_NON_RECENT_DURATION_WEEKS),
            ("5", 5),
        ),
    )
    def test_keep(self, value, expected):
        argv = ["test.db"]
        if value:
            argv.append(value)
        _, args = countme_trim_raw.parse_args(argv)
        assert args.keep == expected


@pytest.mark.parametrize("minmax", ("min", "max"))
def test_get_minmaxtime(minmax):
    connection = mock.Mock()
    cursor = connection.execute.return_value
    result_sentinel = object()
    cursor.fetchone.return_value = (result_sentinel,)

    result = getattr(countme_trim_raw, f"get_{minmax}time")(connection=connection)

    assert result is result_sentinel
    connection.execute.assert_called_once_with(
        f"SELECT {minmax.upper()}(timestamp) FROM countme_raw"
    )
    cursor.fetchone.assert_called_once_with()


@given(value=integers(min_value=constants.COUNTME_START_TIME, max_value=NOW_TIMESTAMP))
def test_next_week(value):
    # Actually use a different algorithm to determine start of next week, special-casing midnights
    # of Mondays, to compare with.
    dt_value = dt.datetime.utcfromtimestamp(value).replace(tzinfo=dt.UTC)
    dt_midnight = dt_value.replace(hour=0, minute=0, second=0, microsecond=0)
    dt_monday = dt_midnight - dt.timedelta(days=dt_value.weekday())
    dt_expected = dt_monday + dt.timedelta(days=7)
    expected = int(dt_expected.timestamp())

    assert countme_trim_raw.next_week(value) == expected


@pytest.mark.parametrize("unique_ip_only", (False, True), ids=("all-entries", "unique-ip-only"))
def test__num_entries(unique_ip_only: bool):
    result_sentinel = object()
    connection = mock.Mock()
    connection.execute.return_value = cursor = mock.Mock()
    cursor.fetchone.return_value = (result_sentinel,)

    result = countme_trim_raw._num_entries(connection, "trim_begin", "trim_end", unique_ip_only)

    assert result is result_sentinel
    expected_query = "SELECT COUNT(*) FROM countme_raw WHERE timestamp >= ? AND timestamp < ?"
    if unique_ip_only:
        expected_query += " AND sys_age < 0"
    connection.execute.assert_called_once_with(expected_query, ("trim_begin", "trim_end"))
    cursor.fetchone.assert_called_once_with()


@pytest.mark.parametrize("unique_ip_only", (False, True), ids=("all-entries", "unique-ip-only"))
def test__del_entries(unique_ip_only):
    connection = mock.Mock()

    countme_trim_raw._del_entries(connection, "trim_begin", "trim_end", unique_ip_only)

    expected_query = "DELETE FROM countme_raw WHERE timestamp >= ? AND timestamp < ?"
    if unique_ip_only:
        expected_query += " AND sys_age < 0"
    connection.execute.assert_called_once_with(expected_query, ("trim_begin", "trim_end"))
    connection.commit.assert_called_once_with()


@given(timestamp=integers(min_value=0, max_value=MAX_TIMESTAMP))
def test_tm2ui(timestamp):
    dt_value = dt.datetime.utcfromtimestamp(timestamp).replace(tzinfo=dt.UTC)
    result = countme_trim_raw.tm2ui(timestamp)
    assert result == dt_value.date().isoformat()


@pytest.mark.parametrize("testcase", ("rw", "rw-unique-ip-only", "rw-interrupt", "ro"))
def test_trim_data(testcase, capsys):
    unique_ip_only = "unique-ip-only" in testcase
    expectation = nullcontext()
    connection = mock.Mock()

    with mock.patch(
        "mirrors_countme.scripts.countme_trim_raw._num_entries"
    ) as _num_entries, mock.patch(
        "mirrors_countme.scripts.countme_trim_raw.tm2ui"
    ) as tm2ui, mock.patch(
        "mirrors_countme.scripts.countme_trim_raw.time"
    ) as time, mock.patch(
        "mirrors_countme.scripts.countme_trim_raw._del_entries"
    ) as _del_entries:
        _num_entries.return_value = "<num_affected>"
        tm2ui.side_effect = lambda v: f"<tm2ui({v})>"
        if "interrupt" in testcase:
            time.sleep.side_effect = KeyboardInterrupt()
            expectation = pytest.raises(KeyboardInterrupt)
        with expectation:
            countme_trim_raw.trim_data(
                connection=connection,
                trim_begin="trim_begin",
                trim_end="trim_end",
                rw="rw" in testcase,
                unique_ip_only=unique_ip_only,
            )

    stdout, _ = capsys.readouterr()

    _num_entries.assert_called_once_with(connection, "trim_begin", "trim_end", unique_ip_only)
    if "rw" in testcase:
        time.sleep.assert_called_once_with(countme_trim_raw.WARN_SECONDS)
        assert "About to DELETE data from <tm2ui(trim_begin)> to <tm2ui(trim_end)>." in stdout
        assert "This will affect <num_affected> entries." in stdout
        assert f"Interrupt within {countme_trim_raw.WARN_SECONDS} seconds to prevent that" in stdout

        if "interrupt" not in testcase:
            assert "DELETING data" in stdout
            _del_entries.assert_called_once_with(
                connection, "trim_begin", "trim_end", unique_ip_only
            )
            assert "Done." in stdout
        else:
            assert "DELETING data" not in stdout
            _del_entries.assert_not_called()
            assert "Done." not in stdout
    else:
        assert "Not deleting data from <tm2ui(trim_begin)> to <tm2ui(trim_end)>." in stdout
        assert "This would affect <num_affected> entries." in stdout
        _del_entries.assert_not_called()


@pytest.mark.parametrize("with_entries", (True, False), ids=("with-entries", "without-entries"))
@pytest.mark.parametrize("unique_ip_only", (False, True), ids=("all-entries", "unique-ip-only"))
@pytest.mark.parametrize("oldest_week", ("without-oldest-week", "with-oldest-week"))
def test_main(oldest_week, unique_ip_only, with_entries, capsys):
    if with_entries:
        expectation = nullcontext()
    else:
        expectation = pytest.raises(SystemExit)

    with mock.patch(
        "mirrors_countme.scripts.countme_trim_raw.parse_args"
    ) as parse_args, mock.patch(
        "mirrors_countme.scripts.countme_trim_raw.sqlite3"
    ) as sqlite3, mock.patch(
        "mirrors_countme.scripts.countme_trim_raw.get_mintime"
    ) as get_mintime, mock.patch(
        "mirrors_countme.scripts.countme_trim_raw.get_maxtime"
    ) as get_maxtime, mock.patch(
        "mirrors_countme.scripts.countme_trim_raw.get_mintime_unique"
    ) as get_mintime_unique, mock.patch(
        "mirrors_countme.scripts.countme_trim_raw.get_maxtime_unique"
    ) as get_maxtime_unique, mock.patch(
        "mirrors_countme.scripts.countme_trim_raw.trim_data"
    ) as trim_data:
        args = mock.Mock(
            sqlite="test.db",
            keep=1,
            oldest_week=oldest_week == "with-oldest-week",
            rw=True,
            unique_ip_only=unique_ip_only,
        )

        parse_args.return_value = (
            object(),  # parser isn’t used
            args,
        )

        sqlite3.connect.return_value = connection = object()

        trim_begin = constants.COUNTME_START_TIME
        if with_entries:
            mintime = trim_begin
        else:
            mintime = None
        get_mintime.return_value = get_mintime_unique.return_value = mintime
        # Act as if there are 4 weeks of data in the DB.
        maxtime = trim_begin + 4 * 7 * 24 * 3600
        get_maxtime.return_value = get_maxtime_unique.return_value = maxtime
        if oldest_week == "without-oldest-week":
            # One week specified to keep plus one week of safety margin, and it’s a float.
            trim_end = float(maxtime - 2 * 7 * 24 * 3600)
        else:
            # This time, it’s an integer. 😁
            trim_end = trim_begin + 7 * 24 * 3600

        with expectation:
            countme_trim_raw._main()

    parse_args.assert_called_once_with()
    sqlite3.connect.assert_called_once_with("file:test.db?mode=rwc", uri=True)
    if unique_ip_only:
        get_mintime.assert_not_called()
        get_maxtime.assert_not_called()
        get_mintime_unique.assert_called_once_with(connection=connection)
        get_maxtime_unique.assert_called_once_with(connection=connection)
    else:
        get_mintime.assert_called_once_with(connection=connection)
        get_maxtime.assert_called_once_with(connection=connection)
        get_mintime_unique.assert_not_called()
        get_maxtime_unique.assert_not_called()

    if with_entries:
        trim_data.assert_called_once_with(
            connection=connection,
            trim_begin=trim_begin,
            trim_end=trim_end,
            rw=True,
            unique_ip_only=unique_ip_only,
        )
    else:
        trim_data.assert_not_called()
        out, err = capsys.readouterr()
        assert "There are no matching entries in the given DB." in out


@pytest.mark.parametrize("interrupt", (False, True))
def test_cli(interrupt):
    with mock.patch("mirrors_countme.scripts.countme_trim_raw._main") as _main:
        if interrupt:
            _main.side_effect = KeyboardInterrupt()
            with pytest.raises(SystemExit) as exc_info:
                countme_trim_raw.cli()
            assert exc_info.value.code == 3
        else:
            countme_trim_raw.cli()

    _main.assert_called_once_with()



from contextlib import nullcontext
from typing import NamedTuple
from unittest import mock

import pytest

from mirrors_countme import readers


class ItemTuple(NamedTuple):
    itemtuple: int


@pytest.fixture
def item_reader_file(tmp_path):
    return str(tmp_path / "test.db")


@pytest.fixture
def item_reader(item_reader_file):
    with mock.patch.object(readers.SQLiteReader, "_get_fields") as _get_fields:
        _get_fields.return_value = ItemTuple._fields
        item_reader = readers.SQLiteReader(item_reader_file, ItemTuple)

    return item_reader


class TestSQLiteReader:
    @pytest.mark.parametrize("testcase", ("happy-path", "no-filefields", "wrong-filefields"))
    def test__init__(self, testcase, item_reader_file):
        if testcase not in ("happy-path", "attr_name"):
            expectation = pytest.raises(readers.ReaderError)
        else:
            expectation = nullcontext()

        with expectation as exc_info, mock.patch.object(
            readers.SQLiteReader, "_get_fields"
        ) as _get_fields:
            match testcase:
                case "happy-path":
                    _get_fields.return_value = ItemTuple._fields
                case "no-filefields":
                    _get_fields.return_value = None
                case "wrong-filefields":
                    _get_fields.return_value = ("all", "the", "wrong", "fields")

            item_reader = readers.SQLiteReader(item_reader_file, ItemTuple, foo="bar")

        match testcase:
            case "happy-path":
                assert item_reader._filename == item_reader_file
                assert item_reader._itemtuple == ItemTuple
                assert item_reader._itemfields == ItemTuple._fields
                assert item_reader._itemfactory == ItemTuple._make
                _get_fields.assert_called_once_with()
            case "no-filefields":
                assert str(exc_info.value) == "no field names found"
            case "wrong-filefields":
                assert str(exc_info.value).startswith("field mismatch:")

    def test__iter_rows(self, item_reader):
        with mock.patch.object(item_reader, "_cursor") as _cursor:
            _cursor.execute.return_value = expected = object()
            result = item_reader._iter_rows()

            assert result is expected
            fields = ",".join(item_reader._itemfields)
            _cursor.execute.assert_called_once_with(
                f"SELECT {fields} FROM {item_reader._tablename}"
            )

    def test___iter__(self, item_reader):
        with mock.patch.object(item_reader, "_iter_rows") as _iter_rows, mock.patch.object(
            item_reader, "_itemfactory"
        ) as _itemfactory:
            iter_rows = [f"iter_row {i}" for i in range(10)]
            _iter_rows.return_value = iter(iter_rows)
            _itemfactory.side_effect = lambda item: f"_itemfactory({item})"

            result = list(iter(item_reader))

            assert result == [f"_itemfactory({item})" for item in iter_rows]

    def test__get_fields(self, item_reader):
        # The item_reader fixture normally shortcuts _get_fields() completely, but we want to be
        # able to test it, so "unmock" the method and only mock out the DB cursor.
        with mock.patch.object(
            item_reader,
            "_get_fields",
            wraps=lambda: readers.SQLiteReader._get_fields(item_reader),
        ), mock.patch.object(item_reader, "_cursor") as _cursor:
            _cursor.execute.return_value = [
                (i, fname) for i, fname in enumerate(item_reader._itemfields)
            ]
            result = item_reader._get_fields()

        assert result == item_reader._itemfields
        _cursor.execute.assert_called_once_with(f"PRAGMA table_info('{item_reader._tablename}')")



# Copyright Red Hat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import csv
import json
import sqlite3

from .util import MinMaxPropMixin

# ===========================================================================
# ====== ItemWriters - output formatting classes ============================
# ===========================================================================


class ItemWriter:
    def __init__(self, fp, itemtuple, timefield="timestamp", **kwargs):
        self._fp = fp
        self._itemtuple = itemtuple
        self._fields = itemtuple._fields
        if timefield not in self._fields:
            raise ValueError(f"{itemtuple.__name__!r} has no time field {timefield!r}")
        self._timefield = timefield
        self._get_writer(**kwargs)

    def _get_writer(self, **kwargs):
        raise NotImplementedError

    def write_item(self, item):
        raise NotImplementedError

    def write_items(self, items):
        for item in items:
            self.write_item(item)

    def write_header(self):
        pass

    def commit(self):
        pass

    def write_index(self):
        pass


class JSONWriter(ItemWriter):
    def _get_writer(self, **kwargs):
        self._dump = json.dump

    def write_item(self, item):
        self._dump(item._asdict(), self._fp)


class CSVWriter(ItemWriter):
    def _get_writer(self, **kwargs):
        self._writer = csv.writer(self._fp)

    def write_header(self):
        self._writer.writerow(self._fields)

    def write_item(self, item):
        self._writer.writerow(item)


class AWKWriter(ItemWriter):
    def _get_writer(self, field_separator="\t", **kwargs):
        self._fieldsep = field_separator

    def _write_row(self, vals):
        self._fp.write(self._fieldsep.join(str(v) for v in vals) + "\n")

    def write_header(self):
        self._write_row(self._fields)

    def write_item(self, item):
        self._write_row(item)


class SQLiteWriter(ItemWriter, MinMaxPropMixin):
    """Write each item as a new row in a SQLite database table."""

    # We have to get a little fancier with types here since SQL tables expect
    # typed values. Good thing Python has types now, eh?
    SQL_TYPE = {
        int: "INTEGER NOT NULL",
        str: "TEXT NOT NULL",
        float: "REAL NOT NULL",
        bytes: "BLOB NOT NULL",
        int | None: "INTEGER",
        str | None: "TEXT",
        float | None: "REAL",
        bytes | None: "BLOB",
    }

    def _sqltype(self, fieldname):
        typehint = self._itemtuple.__annotations__[fieldname]
        return self.SQL_TYPE.get(typehint, "TEXT")

    def _get_writer(self, tablename="countme_raw", **kwargs):
        if hasattr(self._fp, "name"):
            filename = self._fp.name
        else:
            filename = self._fp
        self._connection = sqlite3.connect(f"file:{filename}?mode=rwc", uri=True)
        self._cursor = self._connection.cursor()
        self._tablename = tablename
        self._filename = filename
        # Generate SQL commands so we can use them later.
        # self._create_table creates the table, with column names and types
        # matching the names and types of the fields in self._itemtuple.
        self._create_table = "CREATE TABLE IF NOT EXISTS {table} ({coldefs})".format(
            table=tablename,
            coldefs=",".join(f"{f} {self._sqltype(f)}" for f in self._fields),
        )
        # self._insert_item is an "INSERT" command with '?' placeholders.
        self._insert_item = "INSERT INTO {table} ({colnames}) VALUES ({colvals})".format(
            table=tablename,
            colnames=",".join(self._fields),
            colvals=",".join("?" for f in self._fields),
        )
        # self._create_time_index creates an index on 'timestamp' or whatever
        # the time-series field is.
        self._create_time_index = (
            "CREATE INDEX IF NOT EXISTS {timefield}_idx ON {table} ({timefield})".format(
                table=tablename, timefield=self._timefield
            )
        )

    def write_header(self):
        self._cursor.execute(self._create_table)

    def write_item(self, item):
        self._cursor.execute(self._insert_item, item)

    def write_items(self, items):
        with self._connection:
            self._connection.executemany(self._insert_item, items)

    def commit(self):
        self._connection.commit()

    def write_index(self):
        self._cursor.execute(self._create_time_index)
        self.commit()

    def has_item(self, item):
        """Return True if a row matching `item` exists in this database."""
        condition = " AND ".join(f"{field}=?" for field in self._fields)
        cursor = self._cursor.execute(
            f"SELECT COUNT(*) FROM {self._tablename} WHERE {condition}", item
        )
        return bool(cursor.fetchone()[0])


def make_writer(name, *args, **kwargs):
    """Convenience function to grab/instantiate the right writer"""
    if name == "csv":
        writer = CSVWriter
    elif name == "json":
        writer = JSONWriter
    elif name == "awk":
        writer = AWKWriter
    elif name == "sqlite":
        writer = SQLiteWriter
    else:
        raise ValueError(f"Unknown writer '{name}'")
    return writer(*args, **kwargs)



# Copyright Red Hat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

from typing import NamedTuple

from .util import parse_logtime

# ===========================================================================
# ====== Output items =======================================================
# ===========================================================================


class LogItem(NamedTuple):
    """
    Generic access.log data holder.
    """

    host: str
    identity: str
    time: str
    method: str
    path: str
    query: str | None
    protocol: str
    status: int
    nbytes: int | None
    referrer: str
    user_agent: str

    def datetime(self):
        return parse_logtime(self.time)

    def timestamp(self):
        return parse_logtime(self.time).timestamp()


# TODO: would be kinda nice if there was a clear subclass / translation
# between item classes... or if compile_log_regex made the class for you?
# Or something? It feels like these things should be more closely bound.


class MirrorItem(NamedTuple):
    """
    A basic mirrorlist/metalink metadata item.
    Each item has a timestamp, IP, and the requested repo= and arch= values.
    """

    timestamp: int
    host: str
    repo_tag: str | None
    repo_arch: str | None


class CountmeItem(NamedTuple):
    """
    A "countme" match item.
    Includes the countme value and libdnf User-Agent fields.
    """

    timestamp: int
    host: str
    os_name: str
    os_version: str
    os_variant: str
    os_arch: str
    sys_age: int
    repo_tag: str
    repo_arch: str



import datetime as dt

from hypothesis import given
from hypothesis.strategies import integers

from mirrors_countme import output_items

from .common import MAX_TIMESTAMP


class TestLogItem:
    def make_item(self, dt_value):
        return output_items.LogItem(
            host="127.0.0.1",
            identity="-",
            time=dt_value.strftime("%d/%b/%Y:%H:%M:%S %z"),
            method="GET",
            path="/",
            query=None,
            protocol="https",
            status=200,
            nbytes=None,
            referrer="127.0.0.1",
            user_agent="provocateur",
        )

    @given(timestamp=integers(min_value=0, max_value=MAX_TIMESTAMP))
    def test_datetime(self, timestamp):
        dt_value = dt.datetime.utcfromtimestamp(timestamp).replace(tzinfo=dt.UTC)
        item = self.make_item(dt_value)

        assert item.datetime() == dt_value

    @given(timestamp=integers(min_value=0, max_value=MAX_TIMESTAMP))
    def test_timestamp(self, timestamp):
        dt_value = dt.datetime.utcfromtimestamp(timestamp).replace(tzinfo=dt.UTC)
        item = self.make_item(dt_value)

        assert item.timestamp() == timestamp



from .progress import ReadProgress


def parse_from_iterator(
    logfiles,
    *,
    writer,
    matcher,
    matchmode="countme",
    header=True,
    sqlite=None,
    dupcheck=True,
    index=True,
):
    if header or sqlite:
        writer.write_header()

    if index:
        writer.write_index()

    for logf in logfiles:
        # Make an iterator object for the matching log lines
        match_iter = iter(matcher(logf))

        # TEMP WORKAROUND: filter out match items with missing values
        if matchmode == "countme":
            match_iter = (i for i in match_iter if None not in i)

        if dupcheck:
            # Duplicate data check (for sqlite output)
            for item in match_iter:
                if writer.has_item(item):  # if it's already in the db...
                    continue  # skip to next log

                writer.write_item(item)  # insert it into the db
            writer.commit()
        else:
            # Write matching items (sqlite does commit at end, or rollback on error)
            writer.write_items(match_iter)


def parse(
    *,
    writer,
    matcher,
    matchmode="countme",
    header=True,
    sqlite=None,
    dupcheck=True,
    index=None,
    progress=False,
    logs=None,
):
    parse_from_iterator(
        ReadProgress(logs, display=progress),
        writer=writer,
        matcher=matcher,
        matchmode=matchmode,
        header=header,
        sqlite=sqlite,
        dupcheck=dupcheck,
        index=index,
    )



import datetime
from collections import Counter
from typing import NamedTuple

from .constants import (
    COUNTME_EPOCH,
    COUNTME_EPOCH_ORDINAL,
    COUNTME_START_WEEKNUM,
    LOG_JITTER_WINDOW,
    WEEK_LEN,
)
from .output_items import CountmeItem
from .progress import DIYProgress
from .readers import SQLiteReader
from .util import weeknum
from .writers import CSVWriter, SQLiteWriter

# NOTE: log timestamps do not move monotonically forward, but they don't
# seem to ever jump backwards more than 241 seconds. I assume this is
# some timeout that's set to 4 minutes, and the log entry shows up after
# expiry, or something. Anyway, what this means is that once we've seen
# a timestamp that's 241 seconds past the end of a week, we can assume that
# there will be no further entries whose timestamps belong to the previous
# week.
# We could probably watch the max jitter between log lines and adjust
# this if needed, but for now I'm just gonna pad it to 600 seconds.
# The difference between 241 and 600 is kind of irrelevant - since we get logs
# in 24-hour chunks, any window that extends into the next day means we have to
# wait 24 hours until we can be sure we have all the data for the previous
# week, so the effect would be the same if this was 3600 or 43200 or whatever.


def weekdate(weeknum, weekday=0):
    if weekday < 0 or weekday > 6:
        raise ValueError("weekday must be between 0 (Mon) and 6 (Sun)")
    ordinal = COUNTME_EPOCH_ORDINAL + 7 * int(weeknum) + weekday
    return datetime.date.fromordinal(ordinal)


def daterange(weeknum):
    return weekdate(weeknum, 0), weekdate(weeknum, 6)


# ===========================================================================
# ====== Count Buckets & Items ==============================================
# ===========================================================================


class CountBucket(NamedTuple):
    weeknum: str  # this is a query
    os_name: str
    os_version: str
    os_variant: str
    os_arch: str
    sys_age: str  # this is a key
    repo_tag: str
    repo_arch: str


BucketSelect = CountBucket(
    weeknum=f"((timestamp-{COUNTME_EPOCH})/{WEEK_LEN}) as weeknum",
    os_name="os_name",
    os_version="os_version",
    os_variant="os_variant",
    os_arch="os_arch",
    sys_age="sys_age",
    repo_tag="repo_tag",
    repo_arch="repo_arch",
)


# Same as BucketSelect, but ignore sys_age
BucketSelectUniqueIP = CountBucket(
    weeknum=f"((timestamp-{COUNTME_EPOCH})/{WEEK_LEN}) as weeknum",
    os_name="os_name",
    os_version="os_version",
    os_variant="os_variant",
    os_arch="os_arch",
    sys_age="-1",  # roll all sys_age values into one for totals
    repo_tag="repo_tag",
    repo_arch="repo_arch",
)


class TotalsItem(NamedTuple):
    """TotalsItem is CountBucket with a "hits" count on the front."""

    hits: int
    weeknum: str  # this is a query
    os_name: str
    os_version: str
    os_variant: str
    os_arch: str
    sys_age: str  # this is a key
    repo_tag: str
    repo_arch: str


class CSVCountItem(NamedTuple):
    """
    Represents one row in a countme_totals.csv file.
    In the interest of human-readability, we replace 'weeknum' with the
    start and end dates of that week.
    """

    week_start: str
    week_end: str
    hits: int
    os_name: str
    os_version: str
    os_variant: str
    os_arch: str
    sys_age: int
    repo_tag: str
    repo_arch: str

    @classmethod
    def from_totalsitem(cls, item):
        """Use this method to convert a TotalsItem to a CSVCountItem."""
        hits, weeknum, *rest = item
        week_start, week_end = daterange(weeknum)
        return cls._make([week_start, week_end, hits] + rest)


# ===========================================================================
# ====== SQL + Progress helpers =============================================
# ===========================================================================


class RawDB(SQLiteReader):
    def __init__(self, filename, **kwargs):
        super().__init__(filename, CountmeItem, tablename="countme_raw", **kwargs)

    @property
    def mintime(self):
        return self.mintime_countme

    @property
    def maxtime(self):
        return self.maxtime_countme

    def complete_weeks(self):
        """Compute range of complete weeks in database.

        Return a range(startweek, provweek) that covers (valid + complete)
        weeknums contained in this database. The database may contain some
        data for `provweek`, but since it's provisional/incomplete it's
        outside the range."""
        if self.mintime is None:
            return []
        # startweek can't be earlier than the first week of data
        startweek = max(weeknum(self.mintime), COUNTME_START_WEEKNUM)
        # A week is provisional until the LOG_JITTER_WINDOW expires, so once
        # tsmax minus LOG_JITTER_WINDOW ticks over into a new weeknum, that
        # weeknum is the provisional one. So...
        provweek = max(weeknum(self.maxtime - LOG_JITTER_WINDOW), COUNTME_START_WEEKNUM)
        return range(startweek, provweek)

    def week_iter(self, weeknum, select: tuple | list):
        item_select = ",".join(select)
        start_ts = weeknum * WEEK_LEN + COUNTME_EPOCH
        end_ts = start_ts + WEEK_LEN
        return self._connection.execute(
            f"SELECT {item_select}"
            f" FROM {self._tablename}"
            f" WHERE timestamp >= {start_ts} AND timestamp < {end_ts} AND sys_age >= 0"
        )

    def week_count(self, weeknum):
        cursor = self.week_iter(weeknum, ("COUNT(*)",))
        return cursor.fetchone()[0]


class RawDBU(RawDB):
    def __init__(self, fp, **kwargs):
        super().__init__(fp, **kwargs)

    # As the comment in week_count() says, although we look at both countme
    # and unique IP data ... when we need to know where "the data" starts we
    # only look for unique IP data, otherwise we could process a week that just
    # has countme data as also having unique IP data (and get very low numbers).
    @property
    def mintime(self):
        return self.mintime_unique

    @property
    def maxtime(self):
        return self.maxtime_unique

    def week_iter(self, weeknum, select: tuple | list):
        start_ts = weeknum * WEEK_LEN + COUNTME_EPOCH
        return SplitWeekDays(self, start_ts, select)

    def week_count(self, weeknum):
        start_ts = weeknum * WEEK_LEN + COUNTME_EPOCH
        end_ts = start_ts + WEEK_LEN

        # So this is a "problem" ... we could do a GROUP BY as we do in
        # SplitWeekDays(), but that is basically doing all the same work.
        # Just counting the rows gets us to roughly 8:1 ... so go with that.
        # Also note that we do sys_age < 0, so that we find weeks of data that
        # have unique IP data in it ... even though we'll look at both.
        cursor = self._connection.execute(
            f"SELECT COUNT(*)"
            f" FROM {self._tablename}"
            f" WHERE timestamp >= {start_ts} AND timestamp < {end_ts} AND sys_age < 0"
        )
        return int(cursor.fetchone()[0] / 8)


class SplitWeekDays(object):
    """Pretend to be a SQLite select object. But actually we are 7 of them combined."""

    def __init__(self, rawdb, start_ts, select: tuple | list):
        self.rawdb = rawdb
        self.start_ts = start_ts
        self.select = select

    def __iter__(self):
        return self.fetchall()

    def fetchone(self):
        """Probably don't use this."""
        for x in self.fetchall():
            return x
        else:
            pass  # pragma: no cover

    def fetchall(self):
        """Get a weeks data of unique IPs, by getting group'd data for each day ... for 7 days."""
        item_select = ",".join(self.select)
        start_ts = self.start_ts

        for d in range(7):
            end_ts = start_ts + (WEEK_LEN / 7)
            # Can add "COUNT(*) as nip" to test, but need to remove it for writes
            # Note that we look at _both_ unique/countme data here, so no sys_age
            # checks.
            cursor = self.rawdb._connection.execute(
                f"SELECT {item_select}"
                f" FROM {self.rawdb._tablename}"
                f" WHERE timestamp >= {start_ts} AND timestamp < {end_ts}"
                f" GROUP BY host, os_name, os_version, os_variant, os_arch, repo_tag, repo_arch"
            )
            for row in cursor.fetchall():
                yield row
            start_ts = end_ts


def totals(*, countme_totals, countme_raw=None, progress=False, csv_dump=None):
    # Initialize the writer (better to fail early..)
    totals = SQLiteWriter(
        countme_totals, TotalsItem, timefield="weeknum", tablename="countme_totals"
    )
    totals.write_header()

    # Are we doing an update?
    if countme_raw:
        rawdb = RawDB(countme_raw)

        # Make sure we index them by time.
        totals.write_index()

        # Check to see if there's any new weeks to get data for
        complete_weeks = sorted(rawdb.complete_weeks())
        latest_week_in_totals = totals.maxtime_countme or -1
        new_weeks = [w for w in complete_weeks if w > int(latest_week_in_totals)]

        # Count week by week
        for week in new_weeks:
            # Set up a progress meter and counter
            mon, sun = daterange(week)
            desc = f"week {week} ({mon} -- {sun})"
            if progress:
                total = rawdb.week_count(week)
            else:
                total = 1
            prog = DIYProgress(
                total=total, desc=desc, disable=not progress, unit="row", unit_scale=False
            )
            hitcount = Counter()

            # Select raw items into their buckets and count 'em up
            for bucket in rawdb.week_iter(week, select=BucketSelect):
                hitcount[bucket] += 1
                prog.update()

            # Write the resulting totals into countme_totals
            totals.write_items((hits,) + bucket for bucket, hits in hitcount.items())
            prog.close()

        # Now do roughly the same thing, but for Unique IPs...
        rawdb = RawDBU(countme_raw)

        # Check to see if there's any new weeks to get data for
        complete_weeks = sorted(rawdb.complete_weeks())
        latest_week_in_totals = totals.maxtime_unique or -1
        new_weeks = [w for w in complete_weeks if w > int(latest_week_in_totals)]

        # Count week by week
        for week in new_weeks:
            # Set up a progress meter and counter
            mon, sun = daterange(week)
            desc = f"weeku {week} ({mon} -- {sun})"
            if progress:
                total = rawdb.week_count(week)
            else:
                total = 1
            prog = DIYProgress(
                total=total, desc=desc, disable=not progress, unit="~ip", unit_scale=False
            )
            hitcount = Counter()

            # Select raw items into their buckets and count 'em up
            for bucket in rawdb.week_iter(week, select=BucketSelectUniqueIP):
                hitcount[bucket] += 1
                prog.update()

            # Write the resulting totals into countme_totals
            totals.write_items((hits,) + bucket for bucket, hits in hitcount.items())
            prog.close()
    else:  # pragma: no cover
        pass

    # Was a CSV dump requested?
    if csv_dump:
        totalreader = SQLiteReader(
            countme_totals,
            TotalsItem,
            timefield="weeknum",
            tablename="countme_totals",
        )
        writer = CSVWriter(csv_dump, CSVCountItem, timefield="week_start")
        writer.write_header()
        for item in totalreader:
            writer.write_item(CSVCountItem.from_totalsitem(item))
    else:  # pragma: no cover
        pass



#!/usr/bin/python3
# countme-trim-raw.py - Trim the raw.db file to the next week start.
#
# Copyright Red Hat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: James Antill <jantill@redhat.com>
#
# The main point of this script is to remove all the data from raw.db
# upto the next week start. If you then run it again it'll remove the next week.
# Uses mindate() but then runs direct SQL.

import argparse
import datetime as dt
import locale
import math
import sqlite3
import sys
import time
from functools import partial

from ..constants import COUNTME_EPOCH, WEEK_LEN
from ..util import _get_minmaxtime
from ..version import __version__

CONF_NON_RECENT_DURATION_WEEKS = 13
WARN_SECONDS = 5

locale.setlocale(locale.LC_ALL, "")

# ===========================================================================
# ====== CLI parser & main() ================================================
# ===========================================================================


def positive_int(value):
    value = int(value)
    if value < 1:
        raise ValueError("Value must be > 1")
    return value


def parse_args(argv=None):
    p = argparse.ArgumentParser(
        description="Trim data from raw.db files.",
    )
    p.add_argument("-V", "--version", action="version", version=f"%(prog)s {__version__}")

    p.add_argument(
        "sqlite",
        metavar="DBFILE",
        help="Write to a sqlite database.",
    )
    p.add_argument(
        "--rw",
        "--read-write",
        dest="rw",
        default=False,
        action="store_true",
        help="Actually delete the entries.",
    )
    p.add_argument(
        "--noop",
        dest="rw",
        default=False,
        action="store_false",
        help="Don’t actually delete the entries. (default)",
    )

    p.add_argument(
        "--oldest-week",
        action="store_true",
        help="Trim the oldest (full) week of data.",
    )
    p.add_argument(
        "keep",
        metavar="KEEP",
        type=positive_int,
        nargs="?",
        default=CONF_NON_RECENT_DURATION_WEEKS,
        help=(
            f"Only trim data older than KEEP weeks ago. (default: {CONF_NON_RECENT_DURATION_WEEKS})"
        ),
    )

    p.add_argument("--unique-ip-only", action="store_true", help="Trim only the unique IP data.")

    args = p.parse_args(argv)

    return p, args


get_mintime_unique = partial(
    _get_minmaxtime,
    minmax="min",
    timefield="timestamp",
    tablename="countme_raw",
    where="sys_age < 0",
)
get_maxtime_unique = partial(
    _get_minmaxtime,
    minmax="max",
    timefield="timestamp",
    tablename="countme_raw",
    where="sys_age < 0",
)
get_mintime = partial(_get_minmaxtime, minmax="min", timefield="timestamp", tablename="countme_raw")
get_maxtime = partial(_get_minmaxtime, minmax="max", timefield="timestamp", tablename="countme_raw")


# Find the next week to trim, given the earliest timestamp.
def next_week(mintime: int | float) -> int:
    week_num = math.floor((mintime - COUNTME_EPOCH) / WEEK_LEN) + 1
    return COUNTME_EPOCH + week_num * WEEK_LEN


def _num_entries(
    connection: sqlite3.Connection,
    trim_begin: int | float,
    trim_end: int | float,
    unique_ip_only: bool,
):
    end = ""
    if unique_ip_only:
        end = " AND sys_age < 0"
    cursor = connection.execute(
        "SELECT COUNT(*) FROM countme_raw WHERE timestamp >= ? AND timestamp < ?" + end,
        (trim_begin, trim_end),
    )
    return cursor.fetchone()[0]


def _del_entries(
    connection: sqlite3.Connection,
    trim_begin: int | float,
    trim_end: int | float,
    unique_ip_only: bool,
):
    end = ""
    if unique_ip_only:
        end = " AND sys_age < 0"
    connection.execute(
        "DELETE FROM countme_raw WHERE timestamp >= ? AND timestamp < ?" + end,
        (trim_begin, trim_end),
    )
    connection.commit()


def tm2ui(timestamp):
    date = dt.datetime.fromtimestamp(timestamp, tz=dt.UTC).date()
    return date.isoformat()


def trim_data(
    *,
    connection: sqlite3.Connection,
    trim_begin: int | float,
    trim_end: int | float,
    rw: bool,
    unique_ip_only: bool,
):
    num_affected = _num_entries(connection, trim_begin, trim_end, unique_ip_only)
    if rw:
        print(f" ** About to DELETE data from {tm2ui(trim_begin)} to {tm2ui(trim_end)}. **")
        print(f" ** This will affect {num_affected} entries. **")
        print(f" ** Interrupt within {WARN_SECONDS} seconds to prevent that. **")
        time.sleep(WARN_SECONDS)
        print(" ** DELETING data … **")
        _del_entries(connection, trim_begin, trim_end, unique_ip_only)
        print(" ** Done. **")
    else:
        print(f" ** Not deleting data from {tm2ui(trim_begin)} to {tm2ui(trim_end)}. **")
        print(f" ** This would affect {num_affected} entries. **")


# Real main...
def _main():
    parser, args = parse_args()

    sqlite_uri = f"file:{args.sqlite}?mode=rwc"
    connection = sqlite3.connect(sqlite_uri, uri=True)

    # Find out what timespan is covered by data in database.
    if args.unique_ip_only:
        min_time = get_mintime_unique(connection=connection)
        max_time = get_maxtime_unique(connection=connection)
    else:
        min_time = get_mintime(connection=connection)
        max_time = get_maxtime(connection=connection)

    if min_time is None:
        print(" ** There are no matching entries in the given DB. **")
        sys.exit(0)

    # Calculate last monday midnight UTC in data, to be used for calculating whole week boundaries.
    last = dt.datetime.fromtimestamp(max_time, tz=dt.UTC)
    last_midnight = last.replace(hour=0, minute=0, second=0, microsecond=0)
    last_monday_midnight = last_midnight - dt.timedelta(days=last_midnight.weekday())
    # Add 1 week because the last week contained likely isn’t complete.
    cutoff_monday_midnight = last_monday_midnight - dt.timedelta(weeks=args.keep + 1)

    trim_begin = min_time
    trim_end = max(cutoff_monday_midnight.timestamp(), trim_begin)
    if args.oldest_week:
        trim_end = min(next_week(trim_begin), trim_end)

    trim_data(
        connection=connection,
        trim_begin=trim_begin,
        trim_end=trim_end,
        rw=args.rw,
        unique_ip_only=args.unique_ip_only,
    )


def cli():
    try:
        _main()
    except KeyboardInterrupt:
        raise SystemExit(3)  # sure, 3 is good, why not



import gzip
import lzma
import re
import sys
from contextlib import nullcontext
from unittest import mock

import pytest
from hypothesis import example, given
from hypothesis.strategies import integers

from mirrors_countme import progress

log_line = (
    '121.43.225.226 - - [01/May/2023:00:00:02 +0000] "GET '
    "/metalink?repo=epel-8&arch=x86_64&infra=stock&content=centos&countme=4 "
    'HTTP/1.1" 200 7547 "-" "libdnf (CentOS Linux 8; generic; Linux.x86_64)"'
)

log_line_no_date = (
    r'121.43.225.226 - - "GET /metalink?repo=epel-8&arch=x86_64&infra=stock&'
    r'content=centos&countme=4 HTTP/1.1" 200 7547 "-" "libdnf '
    r'(CentOS Linux 8; generic; Linux.x86_64)"'
)


@pytest.mark.parametrize(
    "date, expected",
    [
        (log_line, "01/May/2023"),
        (log_line_no_date, "??/??/????"),
        (1683208046.7402434, TypeError),
    ],
)
def test_log_date(date, expected):
    if isinstance(expected, str):
        expectation = nullcontext()
    else:
        expectation = pytest.raises(expected)

    with expectation:
        obtained = progress.log_date(date)

    if isinstance(expected, str):
        assert obtained == expected


@pytest.mark.parametrize("filetype", ("plain", "gzip", "xz"))
def test_log_reader(filetype):
    with mock.patch("mirrors_countme.progress.lzma") as lzma, mock.patch(
        "mirrors_countme.progress.gzip"
    ) as gzip, mock.patch("mirrors_countme.progress.open") as open:
        match filetype:
            case "plain":
                fn = open
                filename = "test.log"
            case "gzip":
                fn = gzip.open
                filename = "test.log.gz"
            case "xz":
                fn = lzma.open
                filename = "test.log.xz"

        fn.return_value = result_sentinel = object()

        result = progress.log_reader(filename)

    assert result is result_sentinel
    for open_fn in (lzma.open, gzip.open, open):
        if open_fn is fn:
            open_fn.assert_called_once_with(filename, mode="rt", errors="replace")
        else:
            open_fn.assert_not_called()


@pytest.fixture
def logfile_content() -> bytes:
    return b"x" * 1024


@pytest.fixture
def plain_logfile(logfile_content, tmp_path):
    plain_logfile = tmp_path / "test.log"
    plain_logfile.write_bytes(logfile_content)
    return plain_logfile


@pytest.fixture
def xz_logfile(logfile_content, tmp_path):
    xz_logfile = tmp_path / "test.log.xz"
    with lzma.LZMAFile(xz_logfile, mode="w", format=lzma.FORMAT_XZ) as fp:
        fp.write(logfile_content)
    return xz_logfile


@pytest.fixture
def gz_logfile(logfile_content, tmp_path):
    gz_logfile = tmp_path / "test.log.gz"
    with gzip.GzipFile(gz_logfile, mode="w") as fp:
        fp.write(logfile_content)
    return gz_logfile


@pytest.mark.parametrize("file_exists", (True, False), ids=("file-exists", "file-missing"))
@pytest.mark.parametrize("filetype", ("plain", "gzip", "xz"))
def test_log_total_size(
    filetype, file_exists, logfile_content, plain_logfile, gz_logfile, xz_logfile
):
    with mock.patch(
        "mirrors_countme.progress.xz_log_size", wraps=progress.xz_log_size
    ) as xz_log_size, mock.patch(
        "mirrors_countme.progress.gz_log_size", wraps=progress.gz_log_size
    ) as gz_log_size, mock.patch.object(
        progress.os, "stat", wraps=progress.os.stat
    ) as stat:
        match filetype:
            case "plain":
                filepath = plain_logfile
                wrapped_fn = stat
            case "gzip":
                filepath = gz_logfile
                wrapped_fn = gz_log_size
            case "xz":
                filepath = xz_logfile
                wrapped_fn = xz_log_size

        if not file_exists:
            filepath.unlink()

        result = progress.log_total_size(str(filepath))

    if file_exists:
        assert result == len(logfile_content)
    else:
        assert result is None

    for fn in (stat, gz_log_size, xz_log_size):
        if fn is wrapped_fn:
            fn.assert_called_once_with(str(filepath))
        else:
            fn.assert_not_called()


class TestDIYProgress:
    DISPLAY_TEXT_RE = re.compile(
        r"^Test:\s+(?P<pct>\d+)%\s+\[(?P<hashes_tail>#*[-_=#])(?P<spaces> *)\]\s+"
        + r"(?P<count>\d+(?:\.\d+[kmgtp]?)?b)/"
        + r"(?P<total>\d+(?:\.\d+[kmgtp]?)?b)\s*$"
    )
    TEST_TOTAL = 1000000

    @pytest.mark.parametrize("with_file", (False, True), ids=("without-file", "with-file"))
    def test__init__(self, with_file):
        params = ("desc", "total", "disable", "unit", "unit_scale", "barchar")

        kwargs = {p: p for p in params}
        if with_file:
            kwargs["file"] = "file"

        obj = progress.DIYProgress(**kwargs)

        for param in params:
            assert getattr(obj, param) == param

        if with_file:
            assert obj.file == "file"
        else:
            assert obj.file == sys.stderr

        assert obj.count == 0
        assert obj.showat == 0

    @pytest.mark.parametrize("testcase", ("happy-path", "no-refresh", "disabled"))
    def test_set_description(self, testcase):
        obj = progress.DIYProgress(disable="disabled" in testcase)

        desc_sentinel = object()

        with mock.patch.object(progress.DIYProgress, "display") as display:
            obj.set_description(desc_sentinel, refresh="no-refresh" not in testcase)

        assert obj.desc is desc_sentinel
        if "no-refresh" in testcase or "disabled" in testcase:
            display.assert_not_called()
        else:
            display.assert_called_once_with()

    @pytest.mark.parametrize("testcase", ("tick", "no-tick", "disabled"))
    def test_update(self, testcase):
        obj = progress.DIYProgress(total=self.TEST_TOTAL, disable="disabled" in testcase)

        # update() "ticks" at hundredths and increments before checking
        if "no-tick" in testcase:
            obj.count = self.TEST_TOTAL // 100 - 2
        else:
            obj.count = self.TEST_TOTAL // 100 - 1

        count_prev = obj.count
        obj.showat = showat_prev = self.TEST_TOTAL // 100

        with mock.patch.object(progress.DIYProgress, "display") as display:
            obj.update()

        if "disabled" in testcase:
            # Assert no-op
            assert obj.count == count_prev
            assert obj.showat == showat_prev
            display.assert_not_called()
        else:
            assert obj.count == count_prev + 1
            if "no-tick" in testcase:
                assert obj.showat == showat_prev
                display.assert_not_called()
            else:
                assert obj.showat == showat_prev + self.TEST_TOTAL // 100
                display.assert_called_once_with()

    @given(size=integers(min_value=0, max_value=20))
    def test_iter(self, size):
        obj = progress.DIYProgress()

        with mock.patch.object(progress.DIYProgress, "update") as update:
            result = list(obj.iter(range(size)))

        assert result == list(range(size))
        assert len(update.call_args_list) == size

    @pytest.mark.parametrize("scale", " kmgtpe")
    def test_hrsize(self, scale):
        exp = " kmgtpe".index(scale) * 3
        value = 12.3456 * 10**exp
        result = progress.DIYProgress.hrsize(value)
        if scale == " ":
            scale = ""
        if scale != "e":
            assert result == f"12.3{scale}"
        else:
            assert result == "12345.6p"

    @pytest.mark.parametrize("unit_scale", (True, False), ids=("scaled", "unscaled"))
    @given(count=integers(min_value=0, max_value=int(TEST_TOTAL * 1.01 + 1)))
    @example(count=0)  # 0%
    @example(count=TEST_TOTAL)  # 100%
    @example(count=int(TEST_TOTAL * 1.01 + 1))  # >= 101%
    def test_display(self, unit_scale, count):
        obj = progress.DIYProgress(desc="Test", total=self.TEST_TOTAL, unit_scale=unit_scale)
        obj.count = count

        with mock.patch("mirrors_countme.progress.print") as print, mock.patch.object(
            progress.DIYProgress, "hrsize", wraps=progress.DIYProgress.hrsize
        ) as hrsize:
            obj.display()

        if unit_scale:
            assert hrsize.call_count == 2
        else:
            hrsize.assert_not_called()

        print.assert_called_once()
        assert print.call_args.kwargs == {
            "flush": True,
            "file": obj.file,
            "end": "\r",
        }
        text = print.call_args.args[0]
        match = self.DISPLAY_TEXT_RE.match(text)
        groups = match.groupdict()

        #####################
        # Verify the bar, ugh
        #####################

        pct = (count * 100) // self.TEST_TOTAL
        num_leading_hashes = min(pct, 100) // 4
        expected_tail = "_-=#"[pct % 4]
        num_spaces = 25 - 1 - num_leading_hashes
        assert int(groups["pct"]) == pct
        hashes = groups["hashes_tail"][:-1]
        tail = groups["hashes_tail"][-1]
        if pct < 100:
            assert len(hashes) == num_leading_hashes
            assert tail == expected_tail
            assert len(groups["spaces"]) == num_spaces
        else:
            assert len(hashes) == num_leading_hashes - 1
            assert tail == "#"
            assert len(groups["spaces"]) == num_spaces + 1

        ##########################
        # Verify numbers displayed
        ##########################

        # Cut off trailing "b"
        reported_count = groups["count"][:-1]
        reported_total = groups["total"][:-1]

        if unit_scale:
            assert reported_count == obj.hrsize(count)
            assert reported_total == obj.hrsize(self.TEST_TOTAL)
        else:
            assert reported_count == str(count)
            assert reported_total == str(self.TEST_TOTAL)

    @pytest.mark.parametrize("disable", (False, True), ids=("enabled", "disabled"))
    def test_close(self, disable):
        obj = progress.DIYProgress(disable=disable)

        with mock.patch("mirrors_countme.progress.print") as print:
            obj.close()

        if disable:
            print.assert_not_called()
        else:
            print.assert_called_with(flush=True, file=obj.file)


class TestReadProgress:
    def test___init__(self):
        logs_sentinel = object()
        obj = progress.ReadProgress(logs_sentinel)
        assert obj.logs is logs_sentinel
        assert obj.display

    def test___iter__(self):
        logs = [f"log{i}" for i in range(10)]
        obj = progress.ReadProgress(logs)

        with mock.patch("mirrors_countme.progress.log_reader") as log_reader, mock.patch(
            "mirrors_countme.progress.log_total_size"
        ) as log_total_size, mock.patch.object(
            progress.ReadProgress, "_iter_log_lines"
        ) as _iter_log_lines:
            log_reader.side_effect = lambda logfn: f"log_reader({logfn})"
            log_total_size.side_effect = lambda logfn: f"log_total_size({logfn})"
            _iter_log_lines.side_effect = lambda logf, num, total: (
                f"_iter_log_lines({logf}, {num}, {total})"
            )

            iterables = list(iter(obj))

            assert iterables == [
                f"_iter_log_lines(log_reader(log{i}), {i}, log_total_size(log{i}))"
                for i in range(10)
            ]

    def test__progress_obj(self):
        obj = progress.ReadProgress(object())
        with mock.patch("mirrors_countme.progress.DIYProgress") as DIYProgress:
            DIYProgress.return_value = result_sentinel = object()
            result = obj._progress_obj()
        assert result is result_sentinel
        DIYProgress.assert_called_once_with()

    def test__iter_log_lines(self):
        NUM_LINES = 100
        LOG_LINES = [f"line {i + 1}\n" for i in range(NUM_LINES)]
        TOTAL_SIZE = sum(len(line) for line in LOG_LINES)

        obj = progress.ReadProgress([object()])

        with mock.patch.object(progress.ReadProgress, "_progress_obj") as _progress_obj, mock.patch(
            "mirrors_countme.progress.log_date"
        ) as log_date:
            _progress_obj.return_value = prog = mock.Mock()
            log_date.side_effect = lambda line: f"log_date({line})"

            iterated_lines = list(obj._iter_log_lines(logf=LOG_LINES, num=0, total=TOTAL_SIZE))

        prog.set_description.assert_called_once_with("log 1/1, date=log_date(line 1\n)")
        assert prog.update.call_count == NUM_LINES
        prog.close.assert_called_once_with()

        assert iterated_lines == LOG_LINES



# Copyright Red Hat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import sqlite3
from datetime import datetime, timedelta, timezone
from functools import partial
from typing import Literal
from urllib.parse import parse_qsl

from .constants import COUNTME_EPOCH, MONTHIDX, WEEK_LEN

# ===========================================================================
# ====== Output item definitions and helpers ================================
# ===========================================================================


def weeknum(timestamp):
    return (int(timestamp) - COUNTME_EPOCH) // WEEK_LEN


def offset_to_timezone(offset):
    """Convert a UTC offset like -0400 to a datetime.timezone instance"""
    offmin = 60 * int(offset[1:3]) + int(offset[3:5])
    if offset[0] == "-":
        offmin = -offmin
    return timezone(timedelta(minutes=offmin))


def parse_logtime(logtime):
    # Equivalent to - but faster than - strptime_logtime.
    # It's like ~1.5usec vs 11usec, which might seem trivial but in my tests
    # the regex parser can handle like ~200k lines/sec - or 5usec/line - so
    # an extra ~10usec to parse the time field isn't totally insignificant.
    # (btw, slicing logtime by hand and using re.split are both marginally
    # slower. datetime.fromisoformat is slightly faster but not available
    # in Python 3.6 or earlier.)
    dt, off = logtime.split(" ", 1)
    date, hour, minute, second = dt.split(":", 3)
    day, month, year = date.split("/", 2)
    tz = timezone.utc if off in {"+0000", "-0000"} else offset_to_timezone(off)
    return datetime(
        int(year), MONTHIDX[month], int(day), int(hour), int(minute), int(second), 0, tz
    )


def parse_querydict(querystr):
    """Parse request query the way mirrormanager does (last value wins)"""
    return dict(parse_qsl(querystr, separator="&"))


def _fetchone_or_none(cursor):
    """Return the result, or None, if there was no result. For min/max time."""
    res = cursor.fetchone()
    if res is None:
        return None
    return res[0]


def _get_minmaxtime(
    *,
    connection: sqlite3.Connection | sqlite3.Cursor,
    minmax: Literal["min"] | Literal["max"],
    timefield: str,
    tablename: str,
    where: str | None = None,
):
    query = f"SELECT {minmax.upper()}({timefield}) FROM {tablename}"
    if where:
        query += f" WHERE {where}"
    cursor = connection.execute(query)
    return _fetchone_or_none(cursor)


class MinMaxPropMixin:
    """
    A mixin to add mintime and maxtime properties

    The suffixed properties distinguish between entries for countme and unique
    IP statistics.

    This is for the reader and writer classes which have _cursor, _timefield
    and _tablename attributes.
    """

    def _get_minmaxtime_obj(
        self, minmax: Literal["min"] | Literal["max"], where: str | None = None
    ):
        return _get_minmaxtime(
            connection=self._cursor,
            minmax=minmax,
            timefield=self._timefield,
            tablename=self._tablename,
            where=where,
        )

    _get_mintime = partial(_get_minmaxtime_obj, minmax="min")
    _get_maxtime = partial(_get_minmaxtime_obj, minmax="max")
    _get_mintime_countme = partial(_get_minmaxtime_obj, minmax="min", where="sys_age >= 0")
    _get_maxtime_countme = partial(_get_minmaxtime_obj, minmax="max", where="sys_age >= 0")
    _get_mintime_unique = partial(_get_minmaxtime_obj, minmax="min", where="sys_age < 0")
    _get_maxtime_unique = partial(_get_minmaxtime_obj, minmax="max", where="sys_age < 0")

    mintime = property(_get_mintime)
    maxtime = property(_get_maxtime)
    mintime_countme = property(_get_mintime_countme)
    maxtime_countme = property(_get_maxtime_countme)
    mintime_unique = property(_get_mintime_unique)
    maxtime_unique = property(_get_maxtime_unique)



import datetime
import sqlite3
import tarfile
import tempfile
from pathlib import Path
from typing import Any, List, NamedTuple

import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st

from mirrors_countme.matchers import CountmeMatcher
from mirrors_countme.parse import parse, parse_from_iterator
from mirrors_countme.totals import totals
from mirrors_countme.writers import make_writer

HERE = Path(__file__).parent
TEST_DATA_DIR = HERE.parent / "test_data"


def _test_tarfile_factory(tarfile_path):
    """Wrap tarfile.open() context manager for fixtures

    This attempts to open the tarfile and if successful, extracts its contents
    to the current working directory and yields the Tarfile object. On
    failure, it yields None.
    """
    try:
        tarfp = tarfile.open(tarfile_path, "r:xz")
    except FileNotFoundError:
        yield None
    else:
        with tarfp:
            tarfp.extractall()
            yield tarfp


@pytest.fixture
def log_tar(tmp_path_cwd):
    yield from _test_tarfile_factory(TEST_DATA_DIR / "mirrors.tar.xz")


@pytest.fixture
def db_tar(tmp_path_cwd):
    yield from _test_tarfile_factory(TEST_DATA_DIR / "test_result_cmp.tar.xz")


class Args(NamedTuple):
    writer: Any
    matcher: Any
    dupcheck: bool
    index: Any
    header: bool
    progress: bool
    matchmode: str
    format: str
    sqlite: str
    logs: List[str]


@pytest.fixture
def raw_db_tar(tmp_path_cwd):
    yield from _test_tarfile_factory(TEST_DATA_DIR / "test_result_cmp.tar.xz")


@pytest.fixture
def totals_db_tar(tmp_path_cwd):
    yield from _test_tarfile_factory(TEST_DATA_DIR / "countme_totals.tar.xz")


class ArgsTotal(NamedTuple):
    countme_totals: Any
    countme_raw: Any
    progress: bool
    csv_dump: Any
    sqlite: str


def test_count_totals(tmp_path_cwd, raw_db_tar, totals_db_tar):
    if not raw_db_tar or not totals_db_tar:
        pytest.skip("Test data not found")
    args = ArgsTotal(
        countme_totals=str(tmp_path_cwd / "test_result_totals.db"),
        countme_raw=str(tmp_path_cwd / "test_result_cmp.db"),
        progress=False,
        csv_dump=False,
        sqlite=str(tmp_path_cwd / "test_result_totals"),
    )
    totals(args)
    db = sqlite3.connect(args.sqlite)
    tmp_db = tmp_path_cwd / "countme_totals"
    db.execute(f"ATTACH DATABASE '{tmp_db}' AS test_db;")
    rows_missing = db.execute(
        "select * from test_db.countme_totals except select * from countme_totals;"
    )
    missing = rows_missing.fetchone()
    rows_extra = db.execute(
        "select * from countme_totals except select * from test_db.countme_totals;"
    )
    extra = rows_extra.fetchone()
    assert (
        missing is None and extra is None
    ), f"When comparing db's\n {missing} was missing and\n {extra} was extra"


def test_read_file(tmp_path_cwd, log_tar, db_tar):
    if not log_tar or not db_tar:
        pytest.skip("Test data not found")
    matcher = CountmeMatcher
    args = Args(
        writer=make_writer("sqlite", str(tmp_path_cwd / "test_result.db"), matcher.itemtuple),
        matcher=matcher,
        dupcheck=True,
        index=True,
        header=True,
        progress=False,
        matchmode="countme",
        format="csv",
        sqlite=str(tmp_path_cwd / "test_result.db"),
        logs=[
            str(tmp_path_cwd / "mirrors" / str(i) / "mirrors.fedoraproject.org-access.log")
            for i in range(1, 32)
        ],
    )
    parse(args)
    db = sqlite3.connect(args.sqlite)
    tmp_db = tmp_path_cwd / "test_result_cmp.db"
    db.execute(f"ATTACH DATABASE '{tmp_db}' AS test_db;")
    rows_missing = db.execute("select * from test_db.countme_raw except select * from countme_raw;")
    missing = rows_missing.fetchone()
    rows_extra = db.execute("select * from countme_raw except select * from test_db.countme_raw;")
    extra = rows_extra.fetchone()
    assert (
        missing is None and extra is None
    ), f"When comparing db's\n {missing} was missing and\n {extra} was extra"


def create_logline(ip, date, repo):
    dstr = date.strftime("%d/%b/%Y:%H:%M:%S +0000")
    url = "/metalink?repo=updates-released-f33&arch=x86_64&countme=1"
    agent = "libdnf (Fedora 33; workstation; Linux.x86_64)"
    return f'{ip} - - [{dstr}] "GET {url} HTTP/1.1" 200 32015 "-" "{agent}"'


@st.composite
def log_data(draw):
    ip_sample = st.lists(st.ip_addresses(), min_size=10, max_size=10, unique=True)
    repo = st.sampled_from(["Fedora", "epel-7", "centos8"])
    ips = draw(ip_sample)
    today = datetime.datetime.now()
    dates = [today - datetime.timedelta(days=d, hours=i) for i in range(1, 2) for d in range(1, 14)]

    return sorted(((date, ip, draw(repo)) for ip in ips for date in dates), key=lambda x: x[0])


@settings(suppress_health_check=(HealthCheck.too_slow,), deadline=datetime.timedelta(seconds=1))
@given(loglines=log_data())
def test_log(loglines):
    with tempfile.TemporaryDirectory() as tmp_dir:
        rawdb = f"{tmp_dir}/test.db"
        totalsdb = f"{tmp_dir}/test_generated_totals.db"
        matcher = CountmeMatcher
        parse_from_iterator(
            [(create_logline(ip, date, repo) for date, ip, repo in loglines)],
            matcher=matcher,
            writer=make_writer("sqlite", rawdb, matcher.itemtuple),
            dupcheck=True,
            index=True,
            matchmode="countme",
            sqlite=rawdb,
            header=True,
        )
        db = sqlite3.connect(rawdb)
        rows_no = db.execute("select count(*) from countme_raw;").fetchone()[0]
        assert rows_no == len(loglines)

        csv_dump = open(f"{tmp_dir}/test.csv", "w+")
        totals(
            countme_totals=totalsdb,
            countme_raw=rawdb,
            progress=False,
            csv_dump=csv_dump,
        )
        db = sqlite3.connect(totalsdb)
        rows_no = db.execute("select count(*) from countme_totals;").fetchone()[0]
        assert int(rows_no) > 0

        csv_dump.seek(0)
        assert len(csv_dump.readlines()) > 0



# mirrors_countme.regex - regexes for log matching and parsing
#
# Copyright Red Hat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import re

__all__ = (
    "compile_log_regex",
    "LOG_RE",
    "LIBDNF_USER_AGENT_RE",
    "COUNTME_USER_AGENT_RE",
    "MIRRORS_LOG_RE",
    "COUNTME_LOG_RE",
    "LOG_DATE_RE",
)

# ===========================================================================
# ====== Regexes! Get your regexes here! ====================================
# ===========================================================================

# Log format, according to ansible/roles/httpd/proxy/templates/httpd.conf.j2:
#   LogFormat "%a %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-Agent}i\""
# That's the standard Combined Log Format, with numeric IPs (%a).
#
# Example log line:
#   240.159.140.173 - - [29/Mar/2020:16:04:28 +0000] "GET /metalink?repo=fedora-modular-32&arch=x86_64&countme=1 HTTP/2.0" 200 18336 "-" "libdnf (Fedora 32; workstation; Linux.x86_64)" # noqa
#
# Here it is as a Python regex, with a format placeholder for the actual field
# contents. Default field regexes are in LOG_PATTERN_FIELDS, below, and
# compile_log_regex() lets you construct more interesting regexes that only
# match the lines you care about.
# The request target is split into 'path' and 'query'; 'path' is always
# present but 'query' may be absent, depending on the value of 'query_match'.
# 'query_match' should be '?' (optional), '' (required), or '{0}' (absent).
LOG_PATTERN_FORMAT = (
    r"^"
    r"(?P<host>{host})\s"
    r"(?P<identity>{identity})\s"
    r"(?P<user>{user})\s"
    r"\[(?P<time>{time})\]\s"
    r'"(?P<method>{method})\s'
    r"(?P<path>{path})(?:\?(?P<query>{query})){query_match}"
    r'\s(?P<protocol>{protocol})"\s'
    r"(?P<status>{status})\s"
    r"(?P<nbytes>{nbytes})\s"
    r'"(?P<referrer>{referrer})"\s'
    r'"(?P<user_agent>{user_agent})"\s*'
    r"$"
)

# Pattern for a HTTP header token, as per RFC7230.
# Basically: all printable ASCII chars except '"(),/:;<=>?@[\]{}'
# (see https://tools.ietf.org/html/rfc7230#section-3.2.6)
HTTP_TOKEN_PATTERN = r"[\w\#$%^!&'*+.`|~-]+"

# Here's the default/fallback patterns for each field.
# Note that all fields are non-zero width except query, which is optional,
# and query_match, which should be '?', '', or '{0}', as described above.
LOG_PATTERN_FIELDS = {
    "host": "\\S+",
    "identity": "\\S+",
    "user": "\\S+",
    "time": ".+?",
    "method": HTTP_TOKEN_PATTERN,
    "path": "[^\\s\\?]+",
    "query": "\\S*",
    "query_match": "?",
    "protocol": "HTTP/\\d\\.\\d",
    "status": "\\d+",
    "nbytes": "\\d+|-",
    "referrer": '[^"]+',
    "user_agent": ".+?",
}

# A regex for libdnf/rpm-ostree user-agent strings.
# Examples:
#   "libdnf/0.35.5 (Fedora 32; workstation; Linux.x86_64)"
#   "libdnf (Fedora 32; generic; Linux.x86_64)"
#   "rpm-ostree (Fedora 33; coreos; Linux.x86_64)"
#
# The format, according to libdnf/utils/os-release.cpp:getUserAgent():
#   f"{USER_AGENT} ({os_name} {os_version}; {os_variant}; {os_canon}.{os_arch})"
# where:
#   USER_AGENT = "libdnf" or "libdnf/{LIBDNF_VERSION}" or "rpm-ostree"
#   os_name    = os-release NAME
#   os_version = os-release VERSION_ID
#   os_variant = os-release VARIANT_ID
#   os_canon   = rpm %_os (via libdnf getCanonOS())
#   os_arch    = rpm %_arch (via libdnf getBaseArch())
#
# (libdnf before 0.37.2 used "libdnf/{LIBDNF_VERSION}" as USER_AGENT, but the
# version number was dropped in commit d8d0984 due to privacy concerns.)
#
# For more info on the User-Agent header, see RFC7231, Section 5.5.3:
#   https://tools.ietf.org/html/rfc7231#section-5.5.3)
COUNTME_USER_AGENT_PATTERN = (
    r"(?P<product>(?:libdnf|rpm-ostree)(?:/(?P<product_version>\S+))?)\s+"
    r"\("
    r"(?P<os_name>.*)\s"
    r"(?P<os_version>[0-9a-z._-]*?);\s"
    r"(?P<os_variant>[0-9a-z._-]*);\s"
    r"(?P<os_canon>[\w./]+)\."
    r"(?P<os_arch>\w+)"
    r"\)"
)
COUNTME_USER_AGENT_RE = re.compile(COUNTME_USER_AGENT_PATTERN)
LIBDNF_USER_AGENT_RE = re.compile(COUNTME_USER_AGENT_PATTERN)


def compile_log_regex(flags=0, ascii=True, query_present=None, **kwargs):
    """
    Return a compiled re.Pattern object that should match lines in access_log,
    capturing each field (as listed in LOG_PATTERN_FIELDS) in its own group.

    The default regex to match each field is in LOG_PATTERN_FIELDS but you
    can supply your own custom regexes as keyword arguments, like so:

        mirror_request_pattern = compile_log_regex(path='/foo.*?')

    The `flags` argument is passed to `re.compile()`. Since access_log contents
    should (according to the httpd docs) be ASCII-only, that flag is added by
    default, but you can turn that off by adding 'ascii=False'.

    If `query_present` is True, then the regex only matches lines where the
    target resource has a query string - i.e. query is required.
    If False, it only matches lines *without* a query string.
    If None (the default), the query string is optional.
    """
    if ascii:  # pragma: no branch
        flags |= re.ASCII

    fields = LOG_PATTERN_FIELDS.copy()
    fields.update(kwargs)

    if query_present is not None:
        fields["query_match"] = "" if query_present else "{0}"

    pattern = LOG_PATTERN_FORMAT.format(**fields)

    return re.compile(pattern, flags=flags)


# Default matcher that should match any access.log line
LOG_RE = compile_log_regex()

# Compiled pattern to match all mirrorlist/metalink hits, like mirrorlist.py
MIRRORS_LOG_RE = compile_log_regex(path=r"/metalink|/mirrorlist")

# Compiled pattern for countme lines.
# We only count:
#   * GET requests for /metalink or /mirrorlist,
#   * with libdnf's User-Agent string (see above).
# (We used to count only query strings containing "&countme=\d+", this would not let us gather
# “traditional” unique IP statistics.)
COUNTME_LOG_RE = compile_log_regex(
    method="GET|HEAD",
    query_present=True,
    path=r"/metalink|/mirrorlist",
    status=r"200|302",
    user_agent=COUNTME_USER_AGENT_PATTERN,
)

# Regex for pulling the date out of a log line
LOG_DATE_RE = compile_log_regex(time=r"(?P<date>[^:]+):.*?")



from unittest import mock

import pytest

from mirrors_countme.matchers import CountmeMatcher, MirrorMatcher
from mirrors_countme.scripts import countme_parse_access_log
from mirrors_countme.version import __version__
from mirrors_countme.writers import CSVWriter, SQLiteWriter


@pytest.fixture
def argv():
    return ["--sqlite=raw.db", "a log file"]


class TestParseArgs:
    def test_version(self, capsys):
        with pytest.raises(SystemExit) as excinfo:
            countme_parse_access_log.parse_args(["-V"])

        assert excinfo.value.code == 0
        stdout, _ = capsys.readouterr()
        assert __version__ in stdout

    @pytest.mark.parametrize("missing", ("logfile", "format"))
    def test_missing_mandatory(self, missing, capsys):
        argv = []
        if missing != "logfile":
            argv.append("alogfile")
        if missing != "format":
            argv.extend(("--format", "csv"))

        with pytest.raises(SystemExit):
            countme_parse_access_log.parse_args(argv)

        stdout, stderr = capsys.readouterr()

        assert not stdout

        if missing == "logfile":
            assert "error: the following arguments are required: LOG" in stderr

        if missing == "format":
            assert "error: one of the arguments --sqlite -f/--format is required" in stderr

    def test_logs(self, argv):
        args = countme_parse_access_log.parse_args(argv)
        assert args.logs == [item for item in argv if not item.startswith("-")]

    @pytest.mark.parametrize("option, expected", ((None, False), ("--progress", True)))
    def test_progress(self, option, expected, argv):
        if option:
            argv.append(option)

        args = countme_parse_access_log.parse_args(argv)

        assert bool(args.progress) == expected

    @pytest.mark.parametrize(
        "option, matchmode, matcher",
        (
            (None, "countme", CountmeMatcher),
            ("--matchmode=countme", "countme", CountmeMatcher),
            ("--matchmode=mirrors", "mirrors", MirrorMatcher),
        ),
    )
    def test_matchmode(self, option, matchmode, matcher, argv):
        if matchmode:
            argv.append(option)

        args = countme_parse_access_log.parse_args(argv)

        assert args.matchmode == matchmode
        assert args.matcher == matcher

    def test_mutually_exclusive_group(self, argv, capsys):
        with pytest.raises(SystemExit) as excinfo:
            countme_parse_access_log.parse_args(argv + ["--format=csv"])

        assert excinfo.value.code != 0
        stdout, stderr = capsys.readouterr()
        assert not stdout
        assert "error: argument -f/--format: not allowed with argument --sqlite" in stderr

    @pytest.mark.parametrize(
        "option, modified_flag",
        (("--no-header", "header"), ("--no-index", "index"), ("--no-dup-check", "dupcheck")),
    )
    def test_negating_flags(self, option, modified_flag, argv):
        ALL_FLAGS = ("header", "index", "dupcheck")

        args = countme_parse_access_log.parse_args(argv + [option])

        for flag in ALL_FLAGS:
            if flag == modified_flag:
                assert not getattr(args, flag)
            else:
                assert getattr(args, flag)

    @pytest.mark.parametrize("output_format", ("sqlite", "csv"))
    def test_formats(self, output_format, tmp_path):
        match output_format:
            case "sqlite":
                expected_dupcheck = True
                expected_writer = SQLiteWriter
                rawdb_path = tmp_path / "raw.db"
                argv = [f"--sqlite={rawdb_path}"]
            case "csv":
                expected_dupcheck = False
                argv = ["--format=csv"]
                expected_writer = CSVWriter

        argv.append("a log file")

        args = countme_parse_access_log.parse_args(argv)

        match output_format.split("-")[0]:
            case "sqlite":
                assert args.sqlite.name == str(rawdb_path)
                assert not args.format
            case "csv":
                assert args.format == "csv"
                assert not args.sqlite

        assert args.dupcheck == expected_dupcheck
        assert isinstance(args.writer, expected_writer)


@mock.patch("mirrors_countme.scripts.countme_parse_access_log.parse")
@mock.patch("mirrors_countme.scripts.countme_parse_access_log.parse_args")
def test_cli(parse_args, parse):
    parse_args.return_value = args = mock.Mock()

    countme_parse_access_log.cli()

    parse_args.assert_called_once_with()
    parse.assert_called_once_with(
        matchmode=args.matchmode,
        matcher=args.matcher,
        sqlite=args.sqlite,
        header=args.header,
        index=args.index,
        dupcheck=args.dupcheck,
        writer=args.writer,
        logs=args.logs,
        progress=args.progress,
    )



from collections.abc import Generator
from itertools import chain, repeat
from unittest import mock

import pytest

from mirrors_countme.parse import parse, parse_from_iterator


@pytest.mark.parametrize(
    "testcase",
    (
        "happy-path",
        "without-header",
        "without-countme-workaround",
        "without-dupcheck",
        "without-index",
    ),
)
def test_parse_from_iterator(testcase):
    lines = [
        "not this",
        "this als not",
        "but this (pick me)",
        "and this (pick me, too)",
    ]
    logf = mock.Mock(lines=lines)
    writer = mock.Mock()

    class matcher:
        def __init__(self, logf):
            self.logf = logf

        def __iter__(self):
            for line in logf.lines:
                if "pick me" in line:
                    yield (line,)

    manager = mock.Mock()
    manager.attach_mock(writer.write_header, "write_header")
    manager.attach_mock(writer.write_index, "write_index")
    manager.attach_mock(writer.has_item, "has_item")
    writer.has_item.side_effect = chain([True], repeat(False))
    manager.attach_mock(writer.write_item, "write_item")
    manager.attach_mock(writer.commit, "commit")
    manager.attach_mock(writer.write_items, "write_items")

    header = sqlite = "without-header" not in testcase
    matchmode = "countme" if "without-countme-workaround" not in testcase else "mirrors"
    dupcheck = "without-dupcheck" not in testcase
    index = "without-index" not in testcase

    parse_from_iterator(
        [logf],
        writer=writer,
        matcher=matcher,
        header=header,
        sqlite=sqlite,
        matchmode=matchmode,
        dupcheck=dupcheck,
        index=index,
    )

    expected_calls = []

    if "without-header" not in testcase:
        expected_calls.append(mock.call.write_header())

    if "without-index" not in testcase:
        expected_calls.append(mock.call.write_index())

    if "without-dupcheck" not in testcase:
        had_item = False
        for line in lines:
            if "pick me" in line:
                expected_calls.append(mock.call.has_item((line,)))
                if had_item:
                    expected_calls.append(mock.call.write_item((line,)))
                else:
                    had_item = True
        expected_calls.append(mock.call.commit())
        assert manager.mock_calls == expected_calls
    else:
        mock_calls = manager.mock_calls[:-1]
        assert mock_calls == expected_calls

        # This one is hard to mock precisely, its argument is an on-the-fly-created generator
        last_call = manager.mock_calls[-1]
        assert last_call[0] == "write_items"
        assert len(last_call[1]) == 1
        assert isinstance(last_call[1][0], Generator)
        assert len(last_call[2]) == 0


@pytest.mark.parametrize("use_default", (True, False))
@mock.patch("mirrors_countme.parse.parse_from_iterator")
@mock.patch("mirrors_countme.parse.ReadProgress")
def test_parse(ReadProgress, parse_from_iterator, use_default):
    kwargs = {
        "writer": object(),
        "matcher": object(),
        "header": object(),
        "sqlite": object(),
        "dupcheck": object(),
        "index": object(),
    }

    if use_default:
        expected_kwargs = kwargs.copy()
        expected_kwargs["matchmode"] = "countme"
    else:
        kwargs["matchmode"] = object()
        expected_kwargs = kwargs

    ReadProgress.return_value = read_progress = object()
    expected_args = (read_progress,)

    parse(**kwargs)

    parse_from_iterator.assert_called_once_with(*expected_args, **expected_kwargs)



import datetime as dt
import re
from contextlib import nullcontext
from unittest import mock

import pytest

from mirrors_countme import util
from mirrors_countme.constants import COUNTME_EPOCH


@pytest.mark.parametrize(
    "timestamp, expected",
    [
        (COUNTME_EPOCH, 0),
        (1683208046.7402434, 2782),
        ("1683208046.7402434", ValueError),
    ],
)
def test_weeknum(timestamp, expected):
    if isinstance(expected, int):
        expectation = nullcontext()
    else:
        expectation = pytest.raises(expected)

    with expectation:
        obtained = util.weeknum(timestamp)

    if isinstance(expected, int):
        assert obtained == expected


@pytest.mark.parametrize(
    "offset, expected",
    [
        ("-0400", dt.timezone(dt.timedelta(days=-1, seconds=72000))),
        ("+0500", dt.timezone(dt.timedelta(seconds=18000))),
        ("1683208046.7402434", ValueError),
    ],
)
def test_offset_to_timezone(offset, expected):
    if isinstance(expected, dt.timezone):
        expectation = nullcontext()
    else:
        expectation = pytest.raises(expected)

    with expectation:
        obtained = util.offset_to_timezone(offset)

    if isinstance(expected, dt.timezone):
        assert obtained == expected


@pytest.mark.parametrize(
    "logtime, expected",
    (
        ("29/Mar/2020:16:04:28 +0000", dt.datetime(2020, 3, 29, 16, 4, 28, 0, dt.UTC)),
        ("29/Mar/2020:16:04:28 -0000", dt.datetime(2020, 3, 29, 16, 4, 28, 0, dt.UTC)),
        (
            "29/Mar/2020:16:04:28 +0200",
            dt.datetime(2020, 3, 29, 16, 4, 28, 0, dt.timezone(dt.timedelta(hours=2))),
        ),
        (
            "29/Mar/2020:16:04:28 -0800",
            dt.datetime(2020, 3, 29, 16, 4, 28, 0, dt.timezone(dt.timedelta(hours=-8))),
        ),
    ),
)
def test_parse_logtime(logtime, expected):
    assert util.parse_logtime(logtime) == expected


@pytest.mark.parametrize(
    "querystr, expected",
    (
        ("foo=bar&baz=gnu", {"foo": "bar", "baz": "gnu"}),
        ("foo=bar&foo=baz", {"foo": "baz"}),
    ),
)
def test_parse_querydict(querystr, expected):
    assert util.parse_querydict(querystr) == expected


@pytest.mark.parametrize("has_result", (True, False), ids=("with-result", "without-result"))
def test__fetchone_or_none(has_result):
    cursor = mock.Mock()
    if has_result:
        cursor.fetchone.return_value = [result_sentinel := object()]
    else:
        cursor.fetchone.return_value = None

    result = util._fetchone_or_none(cursor)

    if has_result:
        assert result is result_sentinel
    else:
        assert result is None


class MinMaxProp(util.MinMaxPropMixin):
    def __init__(self):
        self._cursor = mock.Mock()
        self._timefield = "timestamp"
        self._tablename = "countme_raw"


class TestMinMaxPropMixin:
    @pytest.mark.parametrize(
        "property_name",
        (
            "mintime_countme",
            "maxtime_countme",
            "mintime_unique",
            "maxtime_unique",
            "mintime",
            "maxtime",
        ),
    )
    def test_minmaxtime_properties(self, property_name):
        test_obj = MinMaxProp()

        with mock.patch("mirrors_countme.util._fetchone_or_none") as _fetchone_or_none:
            min_or_max = property_name[:3]  # "min" or "max"
            if property_name.endswith("_countme"):
                query_filter_snippet = r"\s+where\s+sys_age\s*>=\s*0"
            elif property_name.endswith("_unique"):
                query_filter_snippet = r"\s+where\s+sys_age\s*<\s*0"
            else:
                query_filter_snippet = ""
            expected_query_re = re.compile(
                rf"^select\s+{min_or_max}\({test_obj._timefield}\)\s+from\s+"
                + rf"{test_obj._tablename}{query_filter_snippet}\s*$",
                re.IGNORECASE,
            )

            _fetchone_or_none.return_value = result_sentinel = object()

            result = getattr(test_obj, property_name)

            assert result is result_sentinel
            test_obj._cursor.execute.assert_called_once()
            assert expected_query_re.match(test_obj._cursor.execute.call_args.args[0])



# mirrors_countme.progress: progress meters for CLI output
#
# Copyright Red Hat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import gzip
import lzma
import os
import subprocess
import sys

from .regex import LOG_DATE_RE

__all__ = (
    "DIYProgress",
    "ReadProgress",
)

# ===========================================================================
# ====== Progress meters & helpers ==========================================
# ===========================================================================


def log_date(line):
    match = LOG_DATE_RE.match(line)
    if match:
        return match["date"]
    return "??/??/????"


def log_reader(logfn):
    if logfn.endswith(".xz"):
        return lzma.open(logfn, mode="rt", errors="replace")
    elif logfn.endswith(".gz"):
        return gzip.open(logfn, mode="rt", errors="replace")
    else:
        return open(logfn, mode="rt", errors="replace")


def xz_log_size(xz_filename):
    cmd = ["xz", "--list", "--robot", xz_filename]
    r = subprocess.run(cmd, stdout=subprocess.PIPE)
    if r.returncode != 0:
        return None
    for line in r.stdout.split(b"\n"):
        f = line.split(b"\t")
        if f[0] == b"totals":
            return int(f[4])
    return None  # pragma: no cover


def gz_log_size(gz_filename):
    cmd = ["gzip", "--quiet", "--list", gz_filename]
    r = subprocess.run(cmd, stdout=subprocess.PIPE)
    if r.returncode != 0:
        return None
    csize, uncsize, ratio, name = r.stdout.split()
    return int(uncsize)


def log_total_size(logfn):
    if logfn.endswith(".xz"):
        return xz_log_size(logfn)
    elif logfn.endswith(".gz"):
        return gz_log_size(logfn)
    else:
        try:
            return os.stat(logfn).st_size
        except Exception:
            return None


class DIYProgress:
    def __init__(
        self,
        desc=None,
        total=None,
        file=None,
        disable=False,
        unit="b",
        unit_scale=True,
        barchar="_-=#",
    ):
        self.desc = desc
        self.total = total
        self.file = sys.stderr if file is None else file
        self.disable = disable
        self.unit = unit
        self.unit_scale = unit_scale
        self.count = 0
        self.showat = 0
        self.barchar = barchar

    def set_description(self, desc=None, refresh=True):
        self.desc = desc
        if refresh and not self.disable:
            self.display()

    def update(self, n=1):
        if self.disable:
            return
        self.count += n
        if self.count >= self.showat:
            self.showat = min(self.total, self.showat + self.total // 100)
            self.display()

    def iter(self, iterable):
        for i in iterable:
            yield i
            self.update()

    @staticmethod
    def hrsize(n):
        for suffix in " kmgtp":
            if n < 1000:
                break
            if suffix != "p":
                n /= 1000
        if suffix == " ":
            suffix = ""
        return f"{n:.1f}{suffix}"

    def display(self):
        unit = self.unit
        desc = self.desc
        if self.unit_scale:
            count = self.hrsize(self.count) + unit
            total = self.hrsize(self.total) + unit
        else:
            count = str(self.count) + unit
            total = str(self.total) + unit
        pct = (self.count * 100) // self.total
        bar = (min(pct, 100) // 4) * self.barchar[-1]
        if pct < 100:
            bar += self.barchar[pct % 4]
        print(
            f"{desc}: {pct:>3}% [{bar:<25}] {count:>7}/{total:<7}",
            flush=True,
            file=self.file,
            end="\r",
        )

    def close(self):
        if self.disable:
            return
        print(flush=True, file=self.file)


class ReadProgress:
    def __init__(self, logs, display=True):
        """logs should be a sequence of line-iterable file-like objects.
        if display is False, no progress output will be printed."""
        self.logs = logs
        self.display = display

    def __iter__(self):
        """Iterator for ReadProgress; yields a sequence of line-iterable
        file-like objects (one for each log in logs)."""
        for num, logfn in enumerate(self.logs):
            logf = log_reader(logfn)
            total = log_total_size(logfn)
            yield self._iter_log_lines(logf, num, total)

    def _progress_obj(self, *args, **kwargs):
        return DIYProgress(*args, **kwargs)

    def _iter_log_lines(self, logf, num, total):
        # Make a progress meter for this file
        prog = self._progress_obj(
            unit="b",
            unit_scale=True,
            total=total,
            disable=not self.display or not total,
        )

        for i, line in enumerate(logf):
            if i == 0:
                # Set log date from first processed line
                desc = f"log {num+1}/{len(self.logs)}, date={log_date(line)}"
                prog.set_description(desc)
            prog.update(len(line))
            yield line
        prog.close()



import datetime as dt
from contextlib import nullcontext
from unittest import mock

import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis.strategies import data, integers, none, one_of

from mirrors_countme import totals, util
from mirrors_countme.constants import (
    COUNTME_EPOCH,
    COUNTME_EPOCH_ORDINAL,
    COUNTME_START_WEEKNUM,
    LOG_JITTER_WINDOW,
    WEEK_LEN,
)


@pytest.mark.parametrize(
    "weeknum, weekday, expected",
    [
        (0, 1, dt.date(1970, 1, 6)),
        (7, 7, ValueError),
    ],
)
def test_weekdate(weeknum, weekday, expected):
    if isinstance(expected, dt.date):
        expectation = nullcontext()
    else:
        expectation = pytest.raises(expected)

    with expectation:
        obtained = totals.weekdate(weeknum, weekday)

    if isinstance(expected, dt.date):
        assert obtained == expected


def test_daterange():
    with mock.patch("mirrors_countme.totals.weekdate") as weekdate:
        TEST_WEEKNUM = 15  # why not
        weekdate.side_effect = lambda weeknum, weekday: (weeknum, weekday)

        assert totals.daterange(weeknum=TEST_WEEKNUM) == ((TEST_WEEKNUM, 0), (TEST_WEEKNUM, 6))
        weekdate.assert_has_calls((mock.call(TEST_WEEKNUM, 0), mock.call(TEST_WEEKNUM, 6)))


class TestCSVCountItem:
    @given(
        weeknum=integers(
            min_value=0,
            max_value=(
                (dt.date(dt.MAXYEAR, 12, 31).toordinal() - dt.date(1970, 1, 1).toordinal()) // 7 - 2
            ),
        )
    )
    def test_from_totalsitem(self, weeknum):
        totals_item = totals.TotalsItem(
            hits=0,
            weeknum=str(weeknum),
            os_name="os_name",
            os_version="os_version",
            os_variant="os_variant",
            os_arch="os_arch",
            sys_age="sys_age",
            repo_tag="repo_tag",
            repo_arch="repo_arch",
        )
        csv_count_item = totals.CSVCountItem.from_totalsitem(totals_item)
        week_start, week_end = totals.daterange(weeknum)
        assert csv_count_item == totals.CSVCountItem(
            week_start=week_start,
            week_end=week_end,
            hits=0,
            os_name="os_name",
            os_version="os_version",
            os_variant="os_variant",
            os_arch="os_arch",
            sys_age="sys_age",
            repo_tag="repo_tag",
            repo_arch="repo_arch",
        )


@pytest.fixture
def rawdb(request, tmp_path):
    cls = request.instance.cls

    with mock.patch.object(cls, "_get_fields") as _get_fields:
        _get_fields.return_value = totals.CountmeItem._fields
        rawdb = cls(str(tmp_path / "raw.db"))

    return rawdb


class TestRawDB:
    cls = totals.RawDB
    minmax_default_prop = "countme"

    def test___init__(self):
        with mock.patch.object(totals.SQLiteReader, "__init__") as super_init:
            obj = self.cls("filename", foo="bar")

        assert isinstance(obj, self.cls)
        super_init.assert_called_once_with(
            "filename", totals.CountmeItem, tablename="countme_raw", foo="bar"
        )

    @pytest.mark.parametrize("propname", ("mintime", "maxtime"))
    def test_minmaxtime(self, propname, rawdb):
        # properties have to be mocked on the class
        with mock.patch.object(
            self.cls, f"{propname}_{self.minmax_default_prop}", new_callable=mock.PropertyMock
        ) as minmaxtime_prop:
            minmaxtime_prop.return_value = sentinel = object()
            assert getattr(rawdb, propname) is sentinel

    @settings(suppress_health_check=(HealthCheck.function_scoped_fixture,))
    @given(data=data())
    def test_complete_weeks(self, data, rawdb):
        mintime = data.draw(one_of(integers(min_value=COUNTME_EPOCH_ORDINAL), none()))
        if mintime is not None:
            maxtime = data.draw(integers(min_value=mintime))
        else:
            maxtime = None

        with mock.patch.object(self.cls, "mintime", new=mintime), mock.patch.object(
            self.cls, "maxtime", new=maxtime
        ):
            result = rawdb.complete_weeks()

        if mintime is None:
            assert result == []
        else:
            assert result.start == max(util.weeknum(mintime), COUNTME_START_WEEKNUM)
            assert result.stop == max(
                util.weeknum(maxtime - LOG_JITTER_WINDOW), COUNTME_START_WEEKNUM
            )

    def test_week_iter(self, rawdb):
        weeknum = 15

        with mock.patch.object(rawdb, "_connection") as _connection:
            _connection.execute.return_value = sentinel = object()
            result = rawdb.week_iter(weeknum, ("COUNT(*)",))

        assert result is sentinel

        start_ts = weeknum * WEEK_LEN + COUNTME_EPOCH
        end_ts = start_ts + WEEK_LEN
        _connection.execute.assert_called_once_with(
            f"SELECT COUNT(*) FROM {rawdb._tablename}"
            + f" WHERE timestamp >= {start_ts} AND timestamp < {end_ts} AND sys_age >= 0"
        )

    def test_week_count(self, rawdb):
        weeknum = 37

        with mock.patch.object(rawdb, "week_iter") as week_iter:
            cursor = week_iter.return_value
            cursor.fetchone.return_value = ("bloop",)
            result = rawdb.week_count(weeknum)

        assert result == "bloop"
        week_iter.assert_called_once_with(weeknum, ("COUNT(*)",))
        cursor.fetchone.assert_called_once_with()


class TestRawDBU(TestRawDB):
    cls = totals.RawDBU
    minmax_default_prop = "unique"

    def test_week_iter(self, rawdb):
        weeknum = 59
        select = totals.BucketSelectUniqueIP

        with mock.patch.object(totals, "SplitWeekDays") as SplitWeekDays:
            SplitWeekDays.return_value = sentinel = object()

            result = rawdb.week_iter(weeknum, select)

        assert result is sentinel
        start_ts = weeknum * WEEK_LEN + COUNTME_EPOCH
        SplitWeekDays.assert_called_once_with(rawdb, start_ts, select)

    def test_week_count(self, rawdb):
        weeknum = 23
        num_entries = 100000

        with mock.patch.object(rawdb, "_connection") as _connection:
            cursor = _connection.execute.return_value
            cursor.fetchone.return_value = (num_entries,)

            result = rawdb.week_count(weeknum)

        assert result == num_entries / 8
        start_ts = weeknum * WEEK_LEN + COUNTME_EPOCH
        end_ts = start_ts + WEEK_LEN
        _connection.execute.assert_called_once_with(
            f"SELECT COUNT(*) FROM {rawdb._tablename}"
            + f" WHERE timestamp >= {start_ts} AND timestamp < {end_ts} AND sys_age < 0"
        )
        cursor.fetchone.assert_called_once_with()


@pytest.fixture
def splitweekdays():
    return totals.SplitWeekDays(mock.Mock(), mock.Mock(), mock.Mock())


class TestSplitWeekDays:
    def test___init__(self):
        rawdb = object()
        start_ts = object()
        select = object()

        swd = totals.SplitWeekDays(rawdb=rawdb, start_ts=start_ts, select=select)

        assert swd.rawdb is rawdb
        assert swd.start_ts is start_ts
        assert swd.select is select

    def test___iter__(self, splitweekdays):
        with mock.patch.object(splitweekdays, "fetchall") as fetchall:
            sentinel = object()
            fetchall.return_value = iter([sentinel])

            result = iter(splitweekdays)

        assert list(result) == [sentinel]
        fetchall.assert_called_once_with()

    def test_fetchone(self, splitweekdays):
        with mock.patch.object(splitweekdays, "fetchall") as fetchall:
            sentinel = object()
            fetchall.return_value = [sentinel, object()]

            result = splitweekdays.fetchone()

        assert result is sentinel
        fetchall.assert_called_once_with()

    def test_fetchall(self, splitweekdays):
        splitweekdays.select = ("foo", "bar")
        splitweekdays.start_ts = 0

        rawdb = splitweekdays.rawdb
        cursors = [mock.Mock() for d in range(7)]
        for d, cursor in enumerate(cursors):
            cursor.fetchall.return_value = [f"entry for day #{d + 1}"]
        rawdb._connection.execute.side_effect = cursors
        rawdb._tablename = "tablename"

        result = list(splitweekdays.fetchall())

        assert result == [f"entry for day #{d + 1}" for d in range(7)]
        day_len = WEEK_LEN / 7
        rawdb._connection.execute.assert_has_calls(
            [
                mock.call(
                    "SELECT foo,bar FROM tablename WHERE"
                    + f" timestamp >= {d * day_len} AND timestamp < {(d + 1) * day_len}"
                    + " GROUP BY"
                    + " host, os_name, os_version, os_variant, os_arch, repo_tag, repo_arch"
                )
            ]
        )
        for cursor in cursors:
            cursor.fetchall.assert_called_once_with()


@pytest.mark.parametrize("with_progress", (True, False), ids=("with-progress", "without-progress"))
@pytest.mark.parametrize("with_csv_dump", (True, False), ids=("with-csv_dump", "without-csv_dump"))
@pytest.mark.parametrize(
    "with_countme_raw", (True, False), ids=("with-countme-raw", "without-countme-raw")
)
def test_totals(with_countme_raw, with_csv_dump, with_progress):
    START_WEEK = 1234
    BUCKET_NUM = 5
    ENTRIES_PER_BUCKET = 3

    with mock.patch.object(totals, "SQLiteWriter") as SQLiteWriter, mock.patch.object(
        totals, "RawDB"
    ) as RawDB, mock.patch.object(totals, "DIYProgress") as DIYProgress, mock.patch.object(
        totals, "RawDBU"
    ) as RawDBU, mock.patch.object(
        totals, "SQLiteReader"
    ) as SQLiteReader, mock.patch.object(
        totals, "CSVWriter"
    ) as CSVWriter:
        totals_db = SQLiteWriter.return_value
        progress = DIYProgress.return_value
        rawdb = RawDB.return_value
        rawdbu = RawDBU.return_value
        csv_writer = CSVWriter.return_value

        totals_db.maxtime_countme = totals_db.maxtime_unique = START_WEEK

        # 1 old, 4 new weeks
        complete_weeks = [START_WEEK + w for w in range(5)]
        rawdb.complete_weeks.return_value = rawdbu.complete_weeks.return_value = complete_weeks
        expected_new_weeks = complete_weeks[1:]

        rawdb.week_iter.side_effect = [
            [
                (f"bucket #{i} week {weeknum} for countme",)
                for i in range(BUCKET_NUM)
                for j in range(ENTRIES_PER_BUCKET)
            ]
            for weeknum in expected_new_weeks
        ]
        rawdbu.week_iter.side_effect = [
            [
                (f"bucket #{i} week {weeknum} for unique",)
                for i in range(BUCKET_NUM)
                for j in range(ENTRIES_PER_BUCKET)
            ]
            for weeknum in expected_new_weeks
        ]
        rawdb.week_count.return_value = rawdbu.week_count.return_value = (
            BUCKET_NUM * ENTRIES_PER_BUCKET
        )

        SQLiteReader.return_value = totalreader_results = [
            totals.TotalsItem(
                hits=week,  # as good as any
                weeknum=str(week),
                os_name="os_name",
                os_version="os_version",
                os_variant="os_variant",
                os_arch="os_arch",
                sys_age="sys_age",
                repo_tag="repo_tag",
                repo_arch="repo_arch",
            )
            for week in expected_new_weeks
        ]

        totals.totals(
            countme_totals="totals.db",
            countme_raw="raw.db" if with_countme_raw else None,
            progress=with_progress,
            csv_dump="totals.csv" if with_csv_dump else None,
        )

        SQLiteWriter.assert_called_once_with(
            "totals.db", totals.TotalsItem, timefield="weeknum", tablename="countme_totals"
        )

        if with_countme_raw:
            totals_db.write_index.assert_called_once_with()
            RawDB.assert_called_once_with("raw.db" if with_countme_raw else None)
            RawDBU.assert_called_once_with("raw.db" if with_countme_raw else None)

            expected_calls = []
            for unique in (False, True):
                for week in expected_new_weeks:
                    start_date = dt.date.fromordinal(totals.COUNTME_EPOCH_ORDINAL) + dt.timedelta(
                        weeks=week
                    )
                    end_date = start_date + dt.timedelta(days=6)
                    start_date_str = start_date.strftime("%F")
                    end_date_str = end_date.strftime("%F")
                    expected_calls.append(
                        mock.call(
                            total=BUCKET_NUM * ENTRIES_PER_BUCKET if with_progress else 1,
                            desc=(
                                f"{'week' if not unique else 'weeku'} {week}"
                                + f" ({start_date_str} -- {end_date_str})"
                            ),
                            disable=not with_progress,
                            unit="row" if not unique else "~ip",
                            unit_scale=False,
                        )
                    )
            assert DIYProgress.call_args_list == expected_calls

            # (RawDB, RawDBU object) * each entry
            assert progress.update.call_count == (
                2 * len(expected_new_weeks) * BUCKET_NUM * ENTRIES_PER_BUCKET
            )
            written_items_by_call = [
                list(call.args[0]) for call in totals_db.write_items.call_args_list
            ]
            assert written_items_by_call == [
                [
                    (ENTRIES_PER_BUCKET, f"bucket #{bucket} week {week} for {what}")
                    for bucket in range(BUCKET_NUM)
                ]
                for what in ("countme", "unique")
                for week in expected_new_weeks
            ]
            assert all(call == mock.call() for call in progress.update.call_args_list)
            assert rawdb.week_iter.call_args_list == [
                mock.call(week, select=totals.BucketSelect) for week in expected_new_weeks
            ]
            assert rawdbu.week_iter.call_args_list == [
                mock.call(week, select=totals.BucketSelectUniqueIP) for week in expected_new_weeks
            ]
        else:
            totals_db.write_index.assert_not_called()
            RawDB.assert_not_called()
            RawDBU.assert_not_called()
            DIYProgress.assert_not_called()

        if with_csv_dump:
            SQLiteReader.assert_called_once_with(
                "totals.db", totals.TotalsItem, timefield="weeknum", tablename="countme_totals"
            )
            CSVWriter.assert_called_once_with(
                "totals.csv", totals.CSVCountItem, timefield="week_start"
            )
            csv_writer.write_header.assert_called_once_with()
            assert csv_writer.write_item.call_args_list == [
                mock.call(totals.CSVCountItem.from_totalsitem(totalsitem))
                for totalsitem in totalreader_results
            ]



DAY_LEN = 24 * 60 * 60
WEEK_LEN = 7 * DAY_LEN
COUNTME_EPOCH = 345600  # =00:00:00 Mon Jan 5 00:00:00 1970 (UTC)
MONTHIDX = {
    "Jan": 1,
    "Feb": 2,
    "Mar": 3,
    "Apr": 4,
    "May": 5,
    "Jun": 6,
    "Jul": 7,
    "Aug": 8,
    "Sep": 9,
    "Oct": 10,
    "Nov": 11,
    "Dec": 12,
}
# TODO: this should probably move into the module somewhere..
LOG_JITTER_WINDOW = 600

# Feb 11 2020 was the date that we branched F32 from Rawhide, so we've decided
# to use that as the starting week for countme data.
COUNTME_START_TIME = 1581292800  # =Mon Feb 10 00:00:00 2020 (UTC)
COUNTME_START_WEEKNUM = 2614

# And here's how you convert a weeknum to a human-readable date
COUNTME_EPOCH_ORDINAL = 719167



# Copyright Red Hat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import sys
from typing import Type

from .output_items import CountmeItem, MirrorItem
from .regex import COUNTME_LOG_RE, MIRRORS_LOG_RE
from .util import parse_logtime, parse_querydict


class LogMatcher:
    """Base class for a LogMatcher, which iterates through a log file"""

    regex = NotImplemented
    itemtuple: Type[MirrorItem] | Type[CountmeItem]

    def __init__(self, fileobj):
        self.fileobj = fileobj

    def iteritems(self):
        # TODO: at this point we're single-threaded and CPU-bound;
        # multithreading would speed things up here.
        for line in self.fileobj:
            match = self.regex.match(line)
            if match:
                try:
                    yield self.make_item(match)
                except Exception:
                    # Paper over any conversion errors
                    print(f"IGNORING MALFORMED LINE: {line!r}", file=sys.stderr)

    __iter__ = iteritems

    @classmethod
    def make_item(cls, match):
        raise NotImplementedError


class MirrorMatcher(LogMatcher):
    """Match all mirrorlist/metalink items, like mirrorlist.py does."""

    regex = MIRRORS_LOG_RE
    itemtuple = MirrorItem

    @classmethod
    def make_item(cls, match):
        timestamp = parse_logtime(match["time"]).timestamp()
        query = parse_querydict(match["query"])
        return cls.itemtuple(
            timestamp=int(timestamp),
            host=match["host"],
            repo_tag=query.get("repo"),
            repo_arch=query.get("arch"),
        )


class CountmeMatcher(LogMatcher):
    """Match the libdnf-style "countme" requests."""

    regex = COUNTME_LOG_RE
    itemtuple = CountmeItem

    @classmethod
    def make_item(cls, match):
        timestamp = parse_logtime(match["time"]).timestamp()
        query = parse_querydict(match["query"])
        return cls.itemtuple(
            timestamp=int(timestamp),
            host=match["host"],
            os_name=match["os_name"],
            os_version=match["os_version"],
            os_variant=match["os_variant"],
            os_arch=match["os_arch"],
            sys_age=int(query.get("countme", -1)),
            repo_tag=query.get("repo"),
            repo_arch=query.get("arch"),
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mirrors_countme', 'mirrors_countme.scripts']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['countme-delete-totals = '
                     'mirrors_countme.scripts.countme_delete_totals:cli',
                     'countme-parse-access-log = '
                     'mirrors_countme.scripts.countme_parse_access_log:cli',
                     'countme-totals = '
                     'mirrors_countme.scripts.countme_totals:cli',
                     'countme-trim-raw = '
                     'mirrors_countme.scripts.countme_trim_raw:cli']}

setup_kwargs = {
    'name': 'mirrors-countme',
    'version': '0.1.2',
    'description': 'Parse access logs and count hosts for DNF mirrors',
    'long_description': '# mirrors-countme\n\nParse http `access_log` data, find DNF `countme` requests, and output\nstructured data that lets us estimate the number of people using various\nFedora releases.\n\nSee [Changes/DNF Better Counting] for more info about the `countme` feature in\ngeneral.\n\n## How it works\n\nThe short version:\n\n* Starting in Fedora 32, DNF adds "countme=N" to one random HTTP request\n  per week for each repo that has its `countme` setting enabled.\n* `parse-access-log.py` parses logs from mirrors.fedoraproject.org, finds\n  those requests, and yields the following information:\n    * request timestamp, repo & arch\n    * client OS name, version, variant, and arch\n    * client "age", from 1-4: 1 week, 1 month, 6 months, or >6 months.\n* We use that data to make cool charts & graphs and estimate how many Fedora\n  users there are and what they\'re using.\n\n## Technical details\n\n### Client behavior & configuration\n\nDNF 4.2.9 added the `countme` option, which [dnf.conf(5)] describes like so:\n\n>    Determines whether a special flag should be added to a single, randomly\n>    chosen metalink/mirrorlist query each week.\n>    This allows the repository owner to estimate the number of systems\n>    consuming it, by counting such queries over a week\'s time, which is much\n>    more accurate than just counting unique IP addresses (which is subject to\n>    both overcounting and undercounting due to short DHCP leases and NAT,\n>    respectively).\n>\n>    The flag is a simple "countme=N" parameter appended to the metalink and\n>    mirrorlist URL, where N is an integer representing the "longevity" bucket\n>    this system belongs to.\n>    The following 4 buckets are defined, based on how many full weeks have\n>    passed since the beginning of the week when this system was installed: 1 =\n>    first week, 2 = first month (2-4 weeks), 3 = six months (5-24 weeks) and 4\n>    = more than six months (> 24 weeks).\n>    This information is meant to help distinguish short-lived installs from\n>    long-term ones, and to gather other statistics about system lifecycle.\n\nNote that the default is False, because we don\'t want to enable this for every\nrepo you have configured.\n\nStarting with Fedora 32, we set `countme=1` in Fedora official repo configs:\n\n```\n[updates]\nname=Fedora $releasever - $basearch - Updates\n#baseurl=http://download.example/pub/fedora/linux/updates/$releasever/Everything/$basearch/\nmetalink=https://mirrors.fedoraproject.org/metalink?repo=updates-released-f$releasever&arch=$basearch\nenabled=1\ncountme=1\nrepo_gpgcheck=0\ntype=rpm\ngpgcheck=1\nmetadata_expire=6h\ngpgkey=file:///etc/pki/rpm-gpg/RPM-GPG-KEY-fedora-$releasever-$basearch\nskip_if_unavailable=False\n```\n\nThis means that the default configuration only adds "countme=N" when using\nofficial Fedora repos, which are all done via HTTPS connections to\nmirrors.fedoraproject.org. "countme=N" does _not_ get added in subsequent\nrequests to the chosen mirror(s).\n\n### Privacy, randomization, and user counting\n\nDNF makes a serious effort to keep the `countme` data anonymous _and_ accurate\nby only sending `countme` with one _random_ request to each enabled repo _per\nweek_. So how does it decide when the week starts, and how does it choose\nwhich request?\n\nFirst, all clients use the same "week": Week 0 started at timestamp 345600\n(Mon 05 Jan 1970 00:00:00 - the first Monday of POSIX time), and weeks are\nexactly 604800 (7&times;24&times;60&times;60) seconds long.\n\nSecond, all clients have the same random chance - currently 1:4 - to send\n`countme` with any request in a given week. Once it\'s been sent, the client\nwon\'t send another `countme` for that repo for the rest of the week.\n\nThe default update interval for the `updates` repo is 6 hours, which means\nthat clients who use `dnf-makecache.service` will probably send `countme`\nsometime in the first 24 hours of a given week - and nothing for the rest of\nthe week.\n\nThis means that _daily_ totals of users are unreliably variable, since the\nstart of the week will have more `countme` requests than the end of the week.\nBut the weekly totals should be a consistent, representative sample of the\ntotal population.\n\nFor more details on how libdnf handles the randomization, see\n[libdnf/repo/Repo.cpp:addCountmeFlag()].\n\n### Data collected\n\nThe only data we look at is in the HTTP request itself. Our log lines are in\nthe standard Combined Log Format, like so[^IPvBeefy]:\n\n```\n240.159.140.173 - - [29/Mar/2020:16:04:28 +0000] "GET /metalink?repo=fedora-modular-32&arch=x86_64&countme=1 HTTP/2.0" 200 18336 "-" "libdnf (Fedora 32; workstation; Linux.x86_64)"\n```\n\n\nWe only look at log lines where the request is "GET", the query string includes\n"countme=N", the result is 200 or 302, and the User-Agent string matches the\nlibdnf User-Agent header.\n\nThe only data we use are the timestamp, the query parameters (`repo`, `arch`,\n`countme`), and the libdnf User-Agent data.\n\n#### libdnf User-Agent data\n\nAs in the log line above, the User-Agent header that libdnf sends looks like this:\n\n```\nUser-Agent: libdnf (Fedora 32; workstation; Linux.x86_64)\n```\n\nThis string is assembled in [libdnf/utils/os-release.cpp:getUserAgent()] and\nthe format is as follows:\n\n```\n{product} ({os_name} {os_version}; {os_variant}; {os_canon}.{os_arch})\n```\n\nwhere the values are:\n\n`product`\n:  "libdnf"\n\n`os_name`\n:  [/etc/os-release] `NAME`\n\n`os_version`\n:  [/etc/os-release] `VERSION_ID`\n\n`os_variant`\n:  [/etc/os-release] `VARIANT_ID`\n\n`os_canon`\n:  rpm `%_os` (via libdnf `getCanonOS()`)\n\n`os_arch`\n:  rpm `%_arch` (via libdnf `getBaseArch()`)\n\n(Older versions of libdnf sent `libdnf/{LIBDNF_VERSION}` for the `product`,\nbut the version string was dropped in libdnf 0.37.2 due to privacy concerns;\nsee [libdnf commit d8d0984].)\n\n#### `repo=`, `arch=`, `countme=`\n\nThe `repo=` and `arch=` values are exactly what\'s set in the URL in the `.repo`\nfile.\n\n`repo` is whatever string appears after `repo=` in the repo\'s `metalink` URL.\nThe values that are accepted for `repo` are determined by [mirrormanager];\nsee [mirrormanager2/lib/repomap.py] for some of the gnarly details there.\n\n`arch` is usually set as `arch=$basearch`, which means that `os_arch` and\n`repo_arch` are usually the same value. But it _is_ valid for a client to\nuse a repo with an `arch=` that\'s different from rpm\'s `%_arch` - for example,\nan i686 system could use an i386 repo - so `repo_arch` and `os_arch` _may_ be\ndifferent values.\n\n`countme`, as discussed in [dnf.conf(5)], is a value from 1 to 4 indicating\nthe "age" of the system, counted in _full_ weeks since the system was\nfirst installed. The values are:\n\n1. One week or less (0-1 weeks)\n2. Up to one month (2-4 weeks)\n3. Up to six months (5-24 weeks)\n4. More than six months (25+ weeks)\n\nThese are defined in [libdnf/repo/Repo.cpp:COUNTME\\_BUCKETS].\n\n\n## OK but how do we actually use it in Fedora?\n\nBecause the raw log data contains IP and timestamps that could be used to\ntrack or identify users, we run the parsing and counting inside private parts\nof the Fedora infrastructure and only publish the anonymous aggregate data.\n\nIn practice, this is a three-part process:\n\n1. Run `countme-update-rawdb.sh` daily to parse log data into `rawdb`\n  * `rawdb` is a SQLite database of structured data for each `countme` hit\n  * Kept private since it contains IP addresses and timestamps\n  * Typical log data: ~6GB/day\n  * Typical parsing time: ~5min (Intel Core i7-6770HQ, 2.60GHz)\n  * Typical rawdb size: ~8MB/day for F32; I\'d guess keeping 1 year of data for\n    3 concurrent releases would take about 10GB.\n  * Retaining historical data lets us quickly recalculate counts if we\n    discover significant errors due to misconfigured/malicious clients\n2. Run `countme-update-totals.sh` to read `rawdb` and update `totalsdb`\n  * Counts up hits for each week, grouped by:\n    * System info: `os_name`, `os_version`, `os_variant`, `os_arch`, `sys_age`\n    * Repo requested: `repo_tag`, `repo_arch`\n  * Only generates data for weeks where we have complete log data\n  * No timestamps or IP addresses\n  * Typical parsing time: small, <=~5s\n  * Typical totalsdb size: ~55KB/week (~700 rows/week) for F32\n  * After update, (re)generate `totals.csv`\n3. Publish updated `totals.db` and `totals.csv`\n  * See https://data-analysis.fedoraproject.org/csv-reports/countme/\n  * Might end up in different places/forms in the future\n  * Can also run countme-weekly-totals-ui.sh command to see text data easily.\n\n## Contributing\n\nYou need to be legally allowed to submit any contribution to this project.\nWhat this means in detail is laid out at the [Developer Certificate of Origin]\nwebsite. The mechanism by which you certify this is adding a `Signed-off-by`\ntrailer to git commit log messages, you can do this by using the\n`--signoff/-s` option to `git commit`.\n\n[^IPvBeefy]: Don\'t worry, 240.159.140.173 is a fake IP address. Actually,\n             it\'s the 4-byte UTF-8 encoding for &#x1f32d;, U+1F32D HOT DOG.\n\n[Changes/DNF Better Counting]: https://fedoraproject.org/wiki/Changes/DNF_Better_Counting\n[dnf.conf(5)]: https://dnf.readthedocs.io/en/latest/conf_ref.html\n[/etc/os-release]: http://man7.org/linux/man-pages/man5/os-release.5.html\n[mirrormanager]: https://github.com/fedora-infra/mirrormanager2\n[mirrormanager2/lib/repomap.py]: https://github.com/fedora-infra/mirrormanager2/blob/master/mirrormanager2/lib/repomap.py\n[libdnf commit d8d0984]: https://github.com/rpm-software-management/libdnf/commit/d8d0984\n[libdnf/utils/os-release.cpp:getUserAgent()]: https://github.com/rpm-software-management/libdnf/blob/0.47.0/libdnf/utils/os-release.cpp#L108\n[libdnf/repo/Repo.cpp:addCountmeFlag()]: https://github.com/rpm-software-management/libdnf/blob/0.47.0/libdnf/repo/Repo.cpp#L1051\n[libdnf/repo/Repo.cpp:COUNTME\\_BUCKETS]: https://github.com/rpm-software-management/libdnf/blob/0.47.0/libdnf/repo/Repo.cpp#L92\n[Developer Certificate of Origin]: https://developercertificate.org\n',
    'author': 'Will Woods',
    'author_email': 'None',
    'maintainer': 'Fedora Infrastructure',
    'maintainer_email': 'admin@fedoraproject.org',
    'url': 'http://github.com/fedora-infra/mirrors-countme',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/python3
# countme-delete-totals.py - Delete the last week from the totals.db file.
#
# Copyright Red Hat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: James Antill <jantill@redhat.com>
#
# The main point of this script is to remove the last weeknum of data from
# totals.db. If you then run it again it'll remove the next week.

import argparse
import locale
import sqlite3
import time

from ..constants import COUNTME_EPOCH, WEEK_LEN
from ..version import __version__

locale.setlocale(locale.LC_ALL, "")

# ===========================================================================
# ====== CLI parser & main() ================================================
# ===========================================================================


def parse_args(argv=None):
    p = argparse.ArgumentParser(
        description="Parse Fedora access.log files.",
    )
    p.add_argument("-V", "--version", action="version", version=f"%(prog)s {__version__}")

    p.add_argument(
        "--sqlite",
        metavar="DBFILE",
        help="write to a sqlite database",
    )
    p.add_argument(
        "--noop",
        dest="rw",
        default=True,
        action="store_false",
        help="Skip deleting the entries.",
    )

    args = p.parse_args(argv)

    return args


def last_week(connection):
    cursor = connection.execute("SELECT MAX(weeknum) FROM countme_totals")
    return cursor.fetchone()[0]


def _num_entries_for(connection, weeknum):
    cursor = connection.execute("SELECT COUNT(*) FROM countme_totals WHERE weeknum = ?", (weeknum,))
    return cursor.fetchone()[0]


def _num_entries(connection):
    cursor = connection.execute("SELECT COUNT(*) FROM countme_totals")
    return cursor.fetchone()[0]


def _del_entries_for(connection, weeknum):
    connection.execute("DELETE FROM countme_totals WHERE weeknum = ?", (weeknum,))
    connection.commit()


def tm2ui(timestamp):
    tm = time.gmtime(timestamp)
    return time.strftime("%Y-%m-%d %H:%M:%S", tm)


def weeknum2tm(weeknum):
    ret = COUNTME_EPOCH
    return ret + int(weeknum) * WEEK_LEN


def num2ui(num):
    ret = locale.format_string("%d", num, grouping=True)
    mlen = len(locale.format_string("%d", 100000000, grouping=True))
    if len(ret) < mlen:
        ret = " " * (mlen - len(ret)) + ret
    return ret


def get_trim_data(sqlite_filename):
    connection = sqlite3.connect(f"file:{sqlite_filename}?mode=rwc", uri=True)
    week = last_week(connection)

    print("Next week     :", week, tm2ui(weeknum2tm(week)))
    print("Entries       :", num2ui(_num_entries(connection)))
    print("Entries to del:", num2ui(_num_entries_for(connection, week)))

    return connection, week


def trim_data(connection, week):
    print(" ** About to DELETE data. **")
    time.sleep(5)
    _del_entries_for(connection, week)


def cli():
    try:
        args = parse_args()
        connection, week = get_trim_data(args.sqlite)
        if args.rw:
            trim_data(connection, week)
    except KeyboardInterrupt:
        raise SystemExit(3)  # sure, 3 is good, why not



from importlib import metadata

from mirrors_countme import version


def test_version_info():
    assert len(version.__version_info__)
    assert all(isinstance(digit, int) for digit in version.__version_info__)
    assert ".".join(str(digit) for digit in version.__version_info__) == metadata.version(
        "mirrors-countme"
    )



import re
from io import StringIO
from typing import NamedTuple
from unittest import mock

import pytest

from mirrors_countme import matchers, output_items


class InTestItem(NamedTuple):
    before: str | None
    after: str | None


class InTestMatcher(matchers.LogMatcher):
    regex = re.compile(r"^(?:\s*(?P<before>\S.*)\s+)?matches(?:\s+(?P<after>\S.*)\s*)?$")
    itemtuple = InTestItem

    @classmethod
    def make_item(cls, match):
        before = match["before"]
        after = match["after"]
        if before and "trip" in before or after and "trip" in after:
            raise ValueError("it made me trip!")
        return cls.itemtuple(before=before, after=after)


class TestLogMatcher:
    def test_iter(self, capsys):
        fileobj = StringIO(
            """
            this matches
            this doesn’t match
            and this matches again
            and this matches but makes make_item() trip
            """
        )
        matcher = InTestMatcher(fileobj)
        matched_items = list(iter(matcher))
        assert matched_items == [
            InTestItem(before="this", after=None),
            InTestItem(before="and this", after="again"),
        ]

        _, stderr = capsys.readouterr()
        assert re.match(
            r"^IGNORING MALFORMED LINE: '\s*and this matches but makes make_item\(\) trip", stderr
        )

    def test_make_item(self):
        with pytest.raises(NotImplementedError):
            matchers.LogMatcher.make_item(object())


class TestMirrorMatcher:
    @mock.patch("mirrors_countme.matchers.parse_querydict")
    @mock.patch("mirrors_countme.matchers.parse_logtime")
    def test_make_item(self, parse_logtime, parse_querydict):
        parse_logtime.return_value = parsed_logtime = mock.Mock()
        parsed_logtime.timestamp.return_value = timestamp = "123"
        parse_querydict.return_value = query = {
            "countme": 1,
            "repo": "the repo",
            "arch": "the repo arch",
        }

        match = {
            "time": "1970-01-01 00:02:03",
            "query": "BOOP",
            "host": "an IP address",
        }

        item = matchers.MirrorMatcher.make_item(match)

        assert isinstance(item, output_items.MirrorItem)

        assert item.timestamp == int(timestamp)
        assert item.host == match["host"]
        assert item.repo_tag == query["repo"]
        assert item.repo_arch == query["arch"]


class TestCountmeMatcher:
    @mock.patch("mirrors_countme.matchers.parse_querydict")
    @mock.patch("mirrors_countme.matchers.parse_logtime")
    def test_make_item(self, parse_logtime, parse_querydict):
        parse_logtime.return_value = parsed_logtime = mock.Mock()
        parsed_logtime.timestamp.return_value = timestamp = "123"
        parse_querydict.return_value = query = {
            "countme": 1,
            "repo": "the repo",
            "arch": "the repo arch",
        }

        match = {
            "time": "1970-01-01 00:02:03",
            "query": "BOOP",
            "host": "an IP address",
            "os_name": "Fedora",
            "os_version": "38",
            "os_variant": "server",
            "os_arch": "x86_64",
            "sys_age": "1",
        }

        item = matchers.CountmeMatcher.make_item(match)

        assert isinstance(item, output_items.CountmeItem)

        for key in ("host", "os_name", "os_version", "os_arch"):
            assert getattr(item, key) == match[key]
        assert item.timestamp == int(timestamp)
        assert item.sys_age == int(query["countme"])
        assert item.repo_tag == query["repo"]
        assert item.repo_arch == query["arch"]



# Copyright Red Hat
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Author: Will Woods <wwoods@redhat.com>

import sqlite3

from .util import MinMaxPropMixin

# ===========================================================================
# ====== SQLiteReader - counterpart to ItemWriter ============================
# ===========================================================================


class ReaderError(RuntimeError):
    pass


class SQLiteReader(MinMaxPropMixin):
    def __init__(
        self, filename, itemtuple, tablename="countme_raw", timefield="timestamp", **kwargs
    ):
        self._filename = filename
        self._itemtuple = itemtuple
        self._itemfields = itemtuple._fields
        self._itemfactory = itemtuple._make
        self._tablename = tablename
        self._timefield = timefield
        # self._connection = sqlite3.connect(f"file:{filename}?mode=ro", uri=True)
        self._connection = sqlite3.connect(self._filename)
        self._cursor = self._connection.cursor()
        filefields = self._get_fields()

        if not filefields:
            raise ReaderError("no field names found")
        if filefields != self._itemfields:
            raise ReaderError(f"field mismatch: expected {self._itemfields}, got {filefields}")

    def _iter_rows(self):
        fields = ",".join(self._itemfields)
        return self._cursor.execute(f"SELECT {fields} FROM {self._tablename}")

    def __iter__(self):
        for item in self._iter_rows():
            yield self._itemfactory(item)

    def _get_fields(self):
        fields_sql = f"PRAGMA table_info('{self._tablename}')"
        filefields = tuple(r[1] for r in self._cursor.execute(fields_sql))
        return filefields



import sqlite3
from random import randint
from unittest import mock

import pytest

from mirrors_countme import constants
from mirrors_countme.scripts import countme_delete_totals


def test_parse_args_with_version_argument():
    argv = ["-V"]
    with pytest.raises(SystemExit) as exc_info:
        countme_delete_totals.parse_args(argv)

    assert exc_info.value.code == 0


def test_parse_args_with_sqlite():
    argv = ["--sqlite", "countme.db"]
    args = countme_delete_totals.parse_args(argv)

    assert args.sqlite == "countme.db"


def test_parse_args_with_noop():
    argv = ["--noop"]
    args = countme_delete_totals.parse_args(argv)

    assert not args.rw


def test_parse_args_with_default_noop():
    argv = []
    args = countme_delete_totals.parse_args(argv)

    assert args.rw


def test_last_week_with_data():
    connection = sqlite3.connect(":memory:")
    cursor = connection.cursor()
    cursor.execute("CREATE TABLE countme_totals (weeknum INT)")
    cursor.execute("INSERT INTO countme_totals VALUES (1), (2), (3)")

    result = countme_delete_totals.last_week(connection)
    assert result == 3


def test_num_entries_for_with_data():
    connection = sqlite3.connect(":memory:")
    cursor = connection.cursor()
    cursor.execute("CREATE TABLE countme_totals (weeknum INT, entry TEXT)")
    cursor.execute("INSERT INTO countme_totals VALUES (1, 'entry1'), (1, 'entry2'), (2, 'entry3')")

    result = countme_delete_totals._num_entries_for(connection, 1)
    assert result == 2


def test_num_entries_for_without_data():
    connection = sqlite3.connect(":memory:")
    cursor = connection.cursor()
    cursor.execute("CREATE TABLE countme_totals (weeknum INT, entry TEXT)")

    result = countme_delete_totals._num_entries_for(connection, 1)
    assert result == 0


def test_num_entries_with_data():
    connection = sqlite3.connect(":memory:")
    cursor = connection.cursor()
    cursor.execute("CREATE TABLE countme_totals (entry TEXT)")
    cursor.execute("INSERT INTO countme_totals VALUES ('entry1'), ('entry2'), ('entry3')")

    result = countme_delete_totals._num_entries(connection)
    assert result == 3


def test_num_entries_without_data():
    connection = sqlite3.connect(":memory:")
    cursor = connection.cursor()
    cursor.execute("CREATE TABLE countme_totals (entry TEXT)")

    result = countme_delete_totals._num_entries(connection)
    assert result == 0


def test_del_entries_for():
    connection = sqlite3.connect(":memory:")
    cursor = connection.cursor()
    cursor.execute("CREATE TABLE countme_totals (weeknum INT, entry TEXT)")
    cursor.execute("INSERT INTO countme_totals VALUES (1, 'entry1'), (1, 'entry2'), (2, 'entry3')")

    countme_delete_totals._del_entries_for(connection, 1)

    cursor.execute("SELECT COUNT(*) FROM countme_totals WHERE weeknum = 1")
    result = cursor.fetchone()[0]
    assert result == 0

    cursor.execute("SELECT COUNT(*) FROM countme_totals WHERE weeknum = 2")
    result = cursor.fetchone()[0]
    assert result == 1

    cursor.execute("SELECT COUNT(*) FROM countme_totals")
    result = cursor.fetchone()[0]
    assert result == 1


def test_tm2ui():
    timestamp = 1687176000

    result = countme_delete_totals.tm2ui(timestamp)

    assert result == "2023-06-19 12:00:00"


def test_weeknum2tm():
    weeknum = randint(1, 53)

    result = countme_delete_totals.weeknum2tm(weeknum)

    expected_timestamp = constants.COUNTME_EPOCH + weeknum * constants.WEEK_LEN

    assert result == expected_timestamp


@pytest.mark.parametrize(
    "number, expected_number",
    (
        (1234567890, "1,234,567,890"),
        (123456789, "123,456,789"),
        (123, "        123"),
    ),
)
def test_num2ui(number, expected_number):
    result = countme_delete_totals.num2ui(number)
    assert result == expected_number


def test_get_trim_data(capsys):
    with mock.patch("mirrors_countme.scripts.countme_delete_totals.sqlite3") as sqlite3, mock.patch(
        "mirrors_countme.scripts.countme_delete_totals.last_week"
    ) as last_week, mock.patch(
        "mirrors_countme.scripts.countme_delete_totals._num_entries"
    ) as _num_entries, mock.patch(
        "mirrors_countme.scripts.countme_delete_totals._num_entries_for"
    ) as _num_entries_for:
        sqlite3.connect.return_value = connection_sentinel = object()
        last_week.return_value = 5
        _num_entries.return_value = 1234
        _num_entries_for.return_value = 123

        connection, week = countme_delete_totals.get_trim_data("test.db")

    stdout, _ = capsys.readouterr()
    assert connection is connection_sentinel
    assert week == 5
    assert "Next week     : 5" in stdout
    assert "Entries       :       1,234" in stdout
    assert "Entries to del:         123" in stdout


def test_trim_data(capsys):
    connection = sqlite3.connect(":memory:")
    cursor = connection.cursor()
    cursor.execute("CREATE TABLE countme_totals (weeknum INT, entry TEXT)")
    cursor.execute("INSERT INTO countme_totals VALUES (1, 'entry1'), (1, 'entry2'), (2, 'entry3')")

    countme_delete_totals.trim_data(connection, 1)

    cursor.execute("SELECT COUNT(*) FROM countme_totals")
    result = cursor.fetchone()[0]
    assert result == 1

    stdout, _ = capsys.readouterr()
    assert " ** About to DELETE data. **" in stdout


@pytest.mark.parametrize("rw", (True, False), ids=("readwrite", "dryrun"))
@mock.patch("mirrors_countme.scripts.countme_delete_totals.trim_data")
@mock.patch("mirrors_countme.scripts.countme_delete_totals.get_trim_data")
@mock.patch("mirrors_countme.scripts.countme_delete_totals.parse_args")
def test_cli(mock_parse_args, mock_get_trim_data, mock_trim_data, rw):
    mock_parse_args.return_value = args = mock.Mock(sqlite="test.db", rw=rw)
    mock_get_trim_data.return_value = (connection := object(), week := object())

    countme_delete_totals.cli()

    mock_parse_args.assert_called_once_with()
    mock_get_trim_data.assert_called_once_with(args.sqlite)

    if rw:
        mock_trim_data.assert_called_once_with(connection, week)
    else:
        mock_trim_data.assert_not_called()


@mock.patch("mirrors_countme.scripts.countme_delete_totals.parse_args")
def test_cli_keyboard_interrupt(mock_parse_args):
    mock_parse_args.side_effect = KeyboardInterrupt

    with pytest.raises(SystemExit) as exc_info:
        countme_delete_totals.cli()
    assert exc_info.value.code == 3
